import numpy as np
import logging
from immuneML.data_model.AIRRSequenceSet import AIRRSequenceSet
from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.protein_embedding.ProteinEmbeddingEncoder import ProteinEmbeddingEncoder
from immuneML.util.NumpyHelper import NumpyHelper
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.Logger import log_memory_usage


class ESMCEncoder(ProteinEmbeddingEncoder):
    """
    Encoder based on a pretrained protein language model by Hayes et al. 2025. The used transformer model is
    "esmc_300m".

    Original publication:
    Hayes, T., Rao, R., Akin, H., Sofroniew, N. J., Oktay, D., Lin, Z., Verkuil, R., Tran, V. Q., Deaton, J.,
    Wiggert, M., Badkundri, R., Shafkat, I., Gong, J., Derry, A., Molina, R. S., Thomas, N., Khan, Y. A.,
    Mishra, C., Kim, C., … Rives, A. (2025). Simulating 500 million years of evolution with a language model.
    Science, 387(6736), 850–858. https://doi.org/10.1126/science.ads0018

    Original GitHub repository with license information: https://github.com/evolutionaryscale/esm

    **Dataset type:**

    - SequenceDatasets

    - ReceptorDatasets

    - RepertoireDatasets

    **Specification arguments:**

    - region_type (RegionType): Which part of the receptor sequence to encode. Defaults to IMGT_CDR3.

    - device (str): Which device to use for model inference - 'cpu', 'cuda', 'mps' - as defined by pytorch.
      Defaults to 'cpu'.

    - num_processes (int): Number of processes to use for parallel processing. Defaults to 1.

    - batch_size (int): The number of sequences to encode at the same time. This could have large impact on memory usage.
      If memory is an issue, try with smaller batch sizes. Defaults to 4096.

    - scale_to_zero_mean (bool): Whether to scale the embeddings to zero mean. Defaults to True.

    - scale_to_unit_variance (bool): Whether to scale the embeddings to unit variance. Defaults to True.

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            encodings:
                my_emsc_encoder:
                    ESMC:
                        region_type: IMGT_CDR3
                        device: cpu
                        num_processes: 4
                        batch_size: 4096

    """

    def __init__(self, name: str = None, region_type: RegionType = RegionType.IMGT_CDR3, device: str = 'cpu',
                 num_processes: int = 1, batch_size: int = 4096, scale_to_zero_mean: bool = True,
                 scale_to_unit_variance: bool = True):
        super().__init__(region_type, name, num_processes, device, batch_size, scale_to_zero_mean=scale_to_zero_mean,
                         scale_to_unit_variance=scale_to_unit_variance)
        self.transformer_link = 'esmc_300m'
        self.embedding_dim = 960

    @staticmethod
    def build_object(dataset: Dataset, **params):
        ParameterValidator.assert_region_type(params, ESMCEncoder.__name__)
        return ESMCEncoder(**{**params, 'region_type': RegionType[params['region_type'].upper()]})

    def _get_model(self, log_location):
        from esm.models.esmc import ESMC

        log_memory_usage(stage="start", location=log_location)
        logging.info(f"ESMC ({self.name}): Loading model: {self.transformer_link}")
        
        model = ESMC.from_pretrained(self.transformer_link)
        log_memory_usage("after model load", log_location)
        
        model = model.to(self.device).eval()
        log_memory_usage("after model to device", log_location)
        
        return model

    def _embed_sequence_set(self, sequence_set: AIRRSequenceSet, seq_field: str) -> np.ndarray:
        import torch

        log_location = f"ESMCEncoder ({self.name})"
        model = self._get_model(log_location)

        sequences = getattr(sequence_set, seq_field)
        sequences = sequences.tolist()
        n_sequences = len(sequences)

        # Create memory-mapped array for embeddings
        embeddings = NumpyHelper.create_memmap_array_in_cache((n_sequences, self.embedding_dim))

        # Process in batches
        for i in range(0, n_sequences, self.batch_size):
            batch_end = min(i + self.batch_size, n_sequences)
            batch = sequences[i:batch_end]
            
            logging.info(
                f"{log_location}: Processing batch {i//self.batch_size + 1}/{(n_sequences-1)//self.batch_size + 1}"
            )

            tokens = model._tokenize(batch)
            sequence_id = tokens != model.tokenizer.pad_token_id

            with torch.no_grad():
                output = model.forward(
                    sequence_tokens=tokens,
                    sequence_id=sequence_id
                )

            batch_embeddings = output.embeddings.cpu().numpy().mean(axis=1)
            embeddings[i:batch_end] = batch_embeddings

            del output, tokens, sequence_id
            if torch.cuda.is_available():
                torch.cuda.empty_cache()

            log_memory_usage(f"after batch {i//self.batch_size + 1}", log_location)

        logging.info(f"{log_location}: Finished processing all sequences")
        return embeddings

    def _get_encoding_name(self) -> str:
        return f"ESMC({self.transformer_link})"

    def _get_model_link(self) -> str:
        return self.transformer_link



import shutil
from pathlib import Path
from zipfile import ZipFile, ZIP_STORED

import numpy as np
import pandas as pd

from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.bnp_util import get_sequence_field_name, write_yaml, read_yaml
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.ElementDataset import SequenceDataset
from immuneML.environment.SequenceType import SequenceType
from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.ml_methods.generative_models.progen.ProGenConfig import ProGenConfig
from immuneML.ml_methods.generative_models.progen.ProGenForCausalLM import ProGenForCausalLM
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder


class ProGen(GenerativeModel):
    """
    ProGen is a transformer-based language model for protein sequences. This class allows fine-tuning of a pre-trained
    ProGen model on immune receptor sequences and generating new sequences. It is based on the ProGen2 implementation
    available at https://github.com/salesforce/progen. It uses the sequences as given in "junction_aa" field in the
    input dataset.

    References:

    Nijkamp, E., Ruffolo, J. A., Weinstein, E. N., Naik, N., & Madani, A. (2023).
    Exploring the boundaries of protein language models. Cell Systems, 14(11), 968–978.e3.
    https://doi.org/10.1016/j.cels.2023.10.002

    **Specification arguments:**

    - locus (str): which locus the sequence come from, e.g., TRB

    - tokenizer_path (Path): path to the ProGen tokenizer file (tokenizer.json)

    - trained_model_path (Path): path to the pre-trained ProGen model directory

    - num_frozen_layers (int): number of transformer layers to freeze during fine-tuning

    - num_epochs (int): number of epochs for fine-tuning

    - learning_rate (float): learning rate for fine-tuning

    - device (str): device to use for training and inference ("cpu" or "cuda")

    - fp16 (bool): whether to use mixed precision training

    - prefix_text (str): text to prepend to each sequence during fine-tuning

    - suffix_text (str): text to append to each sequence during fine-tuning

    - max_new_tokens (int): maximum number of new tokens to generate

    - temperature (float): sampling temperature for sequence generation

    - top_p (float): nucleus sampling parameter for sequence generation

    - prompt (str): prompt text to start the generation

    - num_gen_batches (int): number of batches to split generation into

    - per_device_train_batch_size (int): batch size per device during fine-tuning

    - remove_affixes (bool): whether to remove prefix and suffix from generated sequences

    - seed (int): random seed for reproducibility


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            ml_methods:
                progen_model:
                    ProGen:
                        locus: 'beta'
                        tokenizer_path: '/path/to/tokenizer.json'
                        trained_model_path: '/path/to/pretrained/progen/model'
                        num_frozen_layers: 27
                        num_epochs: 3
                        learning_rate: 0.00004
                        device: 'cuda'
                        fp16: False
                        prefix_text: '<|bos|>1'
                        suffix_text: '2<|eos|>'
                        max_new_tokens: 1024
                        temperature: 1.0
                        top_p: 0.9
                        prompt: '1'
                        num_gen_batches: 1
                        per_device_train_batch_size: 2
                        remove_affixes: True
                        name: 'progen_finetuned_model'
                        region_type: 'IMGT_JUNCTION'
                        seed: 42

    """
    @classmethod
    def load_model(cls, path: Path):
        import torch
        assert path.exists(), f"{cls.__name__}: {path} does not exist."

        model_overview_file = path / 'model_overview.yaml'
        assert model_overview_file.exists(), f"{cls.__name__}: {model_overview_file} is not a file."

        # Uses ProGen weights/tokenizer paths from training. Override in model_overview.yaml if needed.
        model_overview = read_yaml(model_overview_file)
        progen = ProGen(**{k: v for k, v in model_overview.items() if k != 'type'})

        config = ProGenConfig.from_pretrained(path)
        model = ProGenForCausalLM.from_pretrained(path,
                                                  config=config,
                                                  dtype=torch.float32 if not progen.fp16 else torch.float16)

        progen.model = model.to(progen.device).eval()

        return progen

    def __init__(self, locus, tokenizer_path: Path, trained_model_path: Path, num_frozen_layers: int, num_epochs: int,
                 learning_rate: float, device: str, fp16: bool = False, prefix_text: str = "", suffix_text: str = "",
                 max_new_tokens: int = 1024, temperature: float = 1.0, top_p: float = 0.9, prompt: str = "1",
                 num_gen_batches: int = 1, per_device_train_batch_size: int = 2, remove_affixes: bool = True,
                 name: str = None, region_type: str = RegionType.IMGT_JUNCTION.name, seed: int = None, ):
        super().__init__(locus, seed=seed, name=name, region_type=RegionType.get_object(region_type))
        self.sequence_type = SequenceType.AMINO_ACID
        self.tokenizer_path = tokenizer_path
        self.trained_model_path = trained_model_path
        self.num_frozen_layers = num_frozen_layers
        self.num_epochs = num_epochs
        self.learning_rate = learning_rate
        self.device = device  # "cpu" or "cuda"
        self.fp16 = fp16
        self.prefix_text = prefix_text
        self.suffix_text = suffix_text
        self.max_new_tokens = max_new_tokens
        self.temperature = temperature
        self.top_p = top_p
        self.prompt = prompt
        self.num_gen_batches = num_gen_batches
        self.per_device_train_batch_size = per_device_train_batch_size
        self.remove_affixes = remove_affixes
        self.model = None

        from tokenizers import Tokenizer
        from transformers import PreTrainedTokenizerFast

        tokenizer = Tokenizer.from_file(str(self.tokenizer_path))
        self.hf_tokenizer = PreTrainedTokenizerFast(tokenizer_object=tokenizer)
        self.hf_tokenizer.pad_token = "<|pad|>"
        self.hf_tokenizer.eos_token = "<|eos|>"
        self.hf_tokenizer.bos_token = "<|bos|>"

    def fit(self, data, path: Path = None):
        assert path is not None, "ProGen.fit requires a target directory path for training outputs."

        logs_dir, output_dir = self._prepare_training_paths(path)
        tokenized_dataset = self._preprocess_dataset(data)

        from transformers import DataCollatorForLanguageModeling
        data_collator = DataCollatorForLanguageModeling(tokenizer=self.hf_tokenizer, mlm=False)
        config = ProGenConfig.from_pretrained(self.trained_model_path)
        model = ProGenForCausalLM.from_pretrained(self.trained_model_path, config=config)
        self._freeze_model_layers(model)

        from transformers import TrainingArguments
        training_args = TrainingArguments(
            output_dir=str(output_dir),
            per_device_train_batch_size=self.per_device_train_batch_size,
            num_train_epochs=self.num_epochs,
            learning_rate=self.learning_rate,
            fp16=self.fp16,
            use_cpu=True if self.device == "cpu" else False,
            save_safetensors=False,
            logging_dir=str(logs_dir),
            save_total_limit=1,
            save_strategy="no"
        )

        from transformers import Trainer
        trainer = Trainer(
            model=model,
            args=training_args,
            train_dataset=tokenized_dataset,
            tokenizer=self.hf_tokenizer,
            data_collator=data_collator
        )

        print_log(f"{self.name or ProGen.__name__}: starting ProGen fine-tuning.", True)
        trainer.train()
        print_log(f"{self.name or ProGen.__name__}: finished ProGen fine-tuning.", True)
        self.model = trainer.model.to(self.device).eval()

    def _freeze_model_layers(self, model):
        for layer in model.transformer.h[:self.num_frozen_layers]:
            for param in layer.parameters():
                param.requires_grad = False
        for param in model.transformer.ln_f.parameters():
            param.requires_grad = True
        for param in model.lm_head.parameters():
            param.requires_grad = True

    def _preprocess_dataset(self, data):
        from datasets import Dataset as HFDataset
        data_df = data.data.topandas()
        data_df["junction_aa"] = self.prefix_text + data_df["junction_aa"].astype(str).fillna("") + self.suffix_text
        hf_dataset = HFDataset.from_pandas(data_df[["junction_aa"]], preserve_index=False)
        tokenized_dataset = hf_dataset.map(
            self.hf_tokenizer,
            batched=True,
            input_columns="junction_aa",
            fn_kwargs={"truncation": True},
            remove_columns=["junction_aa"],
        )
        return tokenized_dataset

    def _prepare_training_paths(self, path):
        base_path = Path(path)
        base_path.mkdir(parents=True, exist_ok=True)
        output_dir = base_path / "model"
        output_dir.mkdir(parents=True, exist_ok=True)
        logs_dir = base_path / "logs"
        logs_dir.mkdir(parents=True, exist_ok=True)
        return logs_dir, output_dir

    def save_model(self, path: Path) -> Path:
        model_path = PathBuilder.build(path / 'model')
        self.model.save_pretrained(model_path, safe_serialization=False)
        shutil.copy(self.tokenizer_path, model_path / Path(self.tokenizer_path).name)

        skip_export_keys = {"model", "tokenizer", "hf_tokenizer", 'region_type', 'sequence_type'}
        write_yaml(filename=model_path / 'model_overview.yaml',
                   yaml_dict={**{k: v for k, v in vars(self).items() if k not in skip_export_keys},
                              **{'type': self.__class__.__name__,
                                 'locus': self.locus.name}})

        archive_path = path / f"trained_model_{self.name}.zip"
        with ZipFile(archive_path, "w", compression=ZIP_STORED) as archive:
            for file_path in (fp for fp in model_path.rglob("*") if fp.is_file()):
                archive.write(file_path, file_path.relative_to(model_path))

        return archive_path.resolve()

    def generate_sequences(self, count: int, seed: int, path: Path, sequence_type: SequenceType,
                           compute_p_gen: bool) -> Dataset:
        import torch
        prompt_encoding = self.hf_tokenizer(self.prompt, return_tensors="pt")
        prompt_input_ids = prompt_encoding.input_ids.to(self.device)
        prompt_attention_mask = prompt_encoding.attention_mask.to(self.device)
        gen_sequences = []

        num_sequences_per_batch = count // self.num_gen_batches
        for i in range(self.num_gen_batches):
            num_current_sequences = num_sequences_per_batch if i < self.num_gen_batches - 1 else count - (
                    num_sequences_per_batch * (self.num_gen_batches - 1))
            with torch.inference_mode():
                output = self.model.generate(
                    input_ids=prompt_input_ids,
                    attention_mask=prompt_attention_mask,
                    max_new_tokens=self.max_new_tokens,
                    do_sample=True,
                    top_p=self.top_p,
                    temperature=self.temperature,
                    num_return_sequences=num_current_sequences,
                    pad_token_id=self.hf_tokenizer.pad_token_id,
                    return_dict_in_generate=False
                )

            gen_sequences.extend(self.hf_tokenizer.batch_decode(output, skip_special_tokens=True))

            print_log(f"{self.name or ProGen.__name__}: {(i + 1) * num_current_sequences} sequences generated.", True)

        if self.remove_affixes:
            gen_sequences = self._remove_affixes(gen_sequences)

        gen_sequences_df = pd.DataFrame({get_sequence_field_name(self.region_type, self.sequence_type): gen_sequences,
                                         'locus': [self.locus.to_string() for _ in range(count)],
                                         'gen_model_name': [self.name for _ in range(count)]})

        return SequenceDataset.build_from_partial_df(gen_sequences_df, PathBuilder.build(path),
                                                     'synthetic_dataset', {'gen_model_name': [self.name]},
                                                     {'gen_model_name': str})

    def _remove_affixes(self, gen_sequences):
        prefix_text = self.hf_tokenizer.decode(self.hf_tokenizer(self.prefix_text).input_ids,
                                               skip_special_tokens=True)
        suffix_text = self.hf_tokenizer.decode(self.hf_tokenizer(self.suffix_text).input_ids,
                                               skip_special_tokens=True)
        gen_sequences = [seq.replace(prefix_text, '').replace(suffix_text, '') for seq in
                         gen_sequences]
        return gen_sequences

    def is_same(self, model) -> bool:
        raise NotImplementedError

    def compute_p_gens(self, sequences, sequence_type: SequenceType) -> np.ndarray:
        raise RuntimeError

    def compute_p_gen(self, sequence: dict, sequence_type: SequenceType) -> float:
        raise RuntimeError

    def can_compute_p_gens(self) -> bool:
        return False

    def can_generate_from_skewed_gene_models(self) -> bool:
        return False

    def generate_from_skewed_gene_models(self, v_genes: list, j_genes: list, seed: int, path: Path,
                                         sequence_type: SequenceType, batch_size: int, compute_p_gen: bool):
        return RuntimeError



import copy
from pathlib import Path
from typing import List

from immuneML.reports.Report import Report
from immuneML.reports.clustering_method_reports.ClusteringMethodReport import ClusteringMethodReport
from immuneML.reports.clustering_reports.ClusteringReport import ClusteringReport
from immuneML.reports.encoding_reports.EncodingReport import EncodingReport
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.clustering.ClusteringState import ClusteringState
from immuneML.workflows.instructions.clustering.clustering_run_model import ClusteringItem


class ClusteringReportHandler:
    """Manages report generation for clustering results."""

    def __init__(self, reports: List[Report]):
        self.reports = reports

    def run_clustering_reports(self, state: ClusteringState):
        """Generate overall clustering reports."""
        report_path = PathBuilder.build(state.result_path / f'reports/')
        for report in self.reports:
            if isinstance(report, ClusteringReport):
                tmp_report = copy.deepcopy(report)
                tmp_report.result_path = report_path
                tmp_report.state = state
                state.clustering_report_results.append(tmp_report.generate_report())

        if len(self.reports) > 0:
            gen_rep_count = len(state.clustering_report_results)
            print_log(f"{state.config.name}: generated {gen_rep_count} clustering reports.", True)

        return state

    def run_item_reports(self, cl_item: ClusteringItem, analysis_desc: str, run_id: int, path: Path,
                         state: ClusteringState) -> list:
        """Generate reports for individual clustering items."""
        report_path = PathBuilder.build(path / f'reports/')
        report_results = []
        for report in self.reports:
            tmp_report = copy.deepcopy(report)
            tmp_report.result_path = PathBuilder.build(report_path / tmp_report.name)
            if isinstance(report, EncodingReport):
                tmp_report.dataset = cl_item.dataset
                rep_result = tmp_report.generate_report()
                report_results.append(rep_result)
            elif isinstance(report, ClusteringMethodReport):
                tmp_report.item = cl_item
                rep_result = tmp_report.generate_report()
                report_results.append(rep_result)

        if len(self.reports) > 0:
            gen_rep_count = len(report_results)
            print_log(f"{state.config.name}: generated {gen_rep_count} reports for setting "
                      f"{cl_item.cl_setting.get_key()} for {analysis_desc}, run id: {run_id + 1}.", True)

        return report_results



import logging
from pathlib import Path

import pandas as pd
import plotly.express as px

from immuneML.data_model import bnp_util
from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.environment.SequenceType import SequenceType
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.simulation.implants.SeedMotif import SeedMotif
from immuneML.simulation.implants.Signal import Signal
from immuneML.simulation.util.signal_annotation import annotate_sequence_dataset
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class TrueMotifsSummaryBarplot(DataReport):
    """
    This report can be used to show how well motifs (for example, motifs introduced using the Simulation instruction)
    are learned across different generative models. The report shows a bar plot with the proportion of sequences in each
    dataset that contain the given motifs. Bars are grouped by the dataset origin (e.g., train, PWM, VAE, LSTM)
    and the signals provided. The report also shows how many of the sequences are memorized (seen in train data) and
    how many are novel (not seen in train data).

    **Specification arguments:**

    - region_type (str): which part of the sequence to check; e.g., IMGT_CDR3

    - implanted_motifs_per_signal (dict): a nested dictionary that specifies the motif seeds that were implanted in the
      given dataset. The first level of keys in this dictionary represents the different signals. In the inner
      dictionary there should be two keys: "seeds" and "gap_sizes".

      - seeds: a list of motif seeds. The seeds may contain gaps, specified by a '/' symbol.

      - gap_sizes: a list of all the possible gap sizes that were used when implanting a gapped motif seed. When no
        gapped seeds are used, this value has no effect.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_motif_report:
                    TrueMotifsSummaryBarplot:
                        region_type: IMGT_CDR3
                        implanted_motifs_per_signal:
                            my_signal1:
                                seeds:
                                - DEQ
                                gap_sizes:
                                - 0
                            my_signal2:
                                seeds:
                                - AS/G
                                gap_sizes:
                                - 2

    """
    @classmethod
    def build_object(cls, **kwargs):
        location = TrueMotifsSummaryBarplot.__name__
        ParameterValidator.assert_keys_present(list(kwargs.keys()),
                                               ["implanted_motifs_per_signal", "region_type"],
                                               location, "TrueMotifsSummaryBarplot report")
        ParameterValidator.assert_region_type(kwargs, location)
        implanted_motifs_per_signal = kwargs["implanted_motifs_per_signal"]

        ParameterValidator.assert_type_and_value(implanted_motifs_per_signal, dict,
                                                 location, f"implanted_motifs_per_signal")

        for signal in implanted_motifs_per_signal.keys():
            ParameterValidator.assert_type_and_value(implanted_motifs_per_signal[signal], dict, location,
                                                     f"implanted_motifs_per_signal/{signal}")

            ParameterValidator.assert_keys_present(implanted_motifs_per_signal[signal].keys(),
                                                   ["seeds", "gap_sizes"], location,
                                                   f"implanted_motifs_per_signal/{signal}")
            ParameterValidator.assert_type_and_value(implanted_motifs_per_signal[signal]["gap_sizes"], list, location,
                                                     f"implanted_motifs_per_signal/{signal}/gap_sizes")
            ParameterValidator.assert_type_and_value(implanted_motifs_per_signal[signal]["seeds"], list, location,
                                                     f"implanted_motifs_per_signal/{signal}/seeds")
            for gap_size in implanted_motifs_per_signal[signal]["gap_sizes"]:
                ParameterValidator.assert_type_and_value(gap_size, int, location,
                                                         f"implanted_motifs_per_signal/{signal}/gap_sizes",
                                                         min_inclusive=0)
            for seed in implanted_motifs_per_signal[signal]["seeds"]:
                ParameterValidator.assert_type_and_value(seed, str, location,
                                                         f"implanted_motifs_per_signal/{signal}/seeds")

        return TrueMotifsSummaryBarplot(**{**kwargs, 'region_type': RegionType[kwargs['region_type'].upper()]})

    def __init__(self, implanted_motifs_per_signal, dataset: Dataset = None, result_path: Path = None, name: str = None,
                 number_of_processes: int = 1, region_type: RegionType = RegionType.IMGT_CDR3):
        super().__init__(dataset=dataset, result_path=result_path, name=name, number_of_processes=number_of_processes)
        self.implanted_motifs_per_signal = implanted_motifs_per_signal
        self.dataset = dataset
        self.result_path = result_path
        self.region_type = region_type

    def _generate(self):
        PathBuilder.build(self.result_path)

        plot_df = self._get_plotting_data()
        plot_df.to_csv(self.result_path / 'annotated_sequences.csv', index=False)
        report_output_fig = self._plot(plot_df, "true_motifs_summary_barplot")

        return ReportResult(self.name,
                            info="This report shows how well implanted ('ground truth') motifs are recovered by "
                                 "generative models. ",
                            output_tables=[ReportOutput(self.result_path / 'annotated_sequences.csv',
                                                        'annotated sequences')],
                            output_figures=[report_output_fig])

    def _get_plotting_data(self):
        signals = self._get_implanted_signals()
        annotated_sequences = annotate_sequence_dataset(self.dataset, signals=signals, region_type=self.region_type)
        annotated_sequences = self._add_novelty_memorization_label(annotated_sequences)
        return annotated_sequences

    def _add_novelty_memorization_label(self, plotting_data):
        sequence_column = bnp_util.get_sequence_field_name(self.region_type, SequenceType.AMINO_ACID)
        train_set = set(plotting_data.loc[plotting_data['data_origin'] == 'original_train', sequence_column])

        for index, row in plotting_data.iterrows():
            if row['data_origin'] in ['original_train', 'original_test']:
                plotting_data.at[index, 'novelty_memorization'] = 'original'
            else:
                seq = row[sequence_column]
                if seq in train_set:
                    plotting_data.at[index, 'novelty_memorization'] = 'memorized'
                else:
                    plotting_data.at[index, 'novelty_memorization'] = 'novel'

        return plotting_data

    @staticmethod
    def _get_sequence_counts(plotting_data, signal_names):
        signal_mask = plotting_data[signal_names].sum(axis=1) > 0
        seq_count_with_signals = plotting_data[signal_mask].groupby('data_origin').size().reset_index(
                                                                        name='seq_count_with_signal')

        seq_count_total = plotting_data.groupby('data_origin').size().reset_index(name='total_sequences')
        seq_counts_df = seq_count_with_signals.merge(seq_count_total, on='data_origin', how='left')
        seq_counts_df['signal_specific_percent'] = seq_counts_df['seq_count_with_signal'] / seq_counts_df['total_sequences'] * 100
        seq_counts_df['label'] = seq_counts_df.apply(
            lambda row: f"{row['data_origin']}<br>Signal-specific sequences: {row['signal_specific_percent']:.2f}%", axis=1)

        return seq_counts_df

    @staticmethod
    def _prepare_and_sort_plotting_data(plotting_data, signal_names, sorted_data_origins, seq_counts_df):
        df_grouped = plotting_data.groupby(['data_origin', 'novelty_memorization'])[signal_names].sum().reset_index()
        df_grouped['total_count'] = df_grouped[signal_names].sum(axis=1)

        df_melted = df_grouped.drop(columns='total_count').melt(id_vars=['data_origin', 'novelty_memorization'],
                                                                var_name='signal',
                                                                value_name='count')
        df_melted['data_origin'] = pd.Categorical(df_melted['data_origin'],
                                                  categories=sorted_data_origins,
                                                  ordered=True)

        df_melted = df_melted.sort_values(['data_origin', 'novelty_memorization', 'signal'])
        df_melted = df_melted.merge(seq_counts_df, on='data_origin', how='left')
        df_melted['frequency'] = df_melted.groupby(['data_origin', 'signal'])['count'].transform(
            lambda x: x / df_melted['total_sequences'])

        return df_melted

    def _plot(self, plotting_data, output_name):
        if plotting_data.empty:
            logging.warning(f"Empty data, skipping {output_name} plot...")
        else:
            filename = self.result_path / f"{output_name}.html"

            plotting_data = plotting_data[plotting_data['data_origin'] != 'original_test']
            signal_names = list(self.implanted_motifs_per_signal.keys())

            seq_counts_df = self._get_sequence_counts(plotting_data, signal_names)
            seq_counts_df['sort_key'] = seq_counts_df['data_origin'].apply(
                lambda x: (0, 0) if x == 'original_train' else (
                    1, -seq_counts_df.loc[seq_counts_df['data_origin'] == x, 'signal_specific_percent'].iloc[0]))
            sorted_data_origins = seq_counts_df.sort_values('sort_key')['data_origin'].tolist()

            df_melted = self._prepare_and_sort_plotting_data(plotting_data, signal_names, sorted_data_origins,
                                                             seq_counts_df)
            df_melted['data_origin_novelty'] = df_melted['data_origin'] + ' - ' + df_melted['novelty_memorization']

            figure = px.bar(
                df_melted,
                x='signal',
                y='frequency',
                color='data_origin_novelty',
                facet_col='label',
                color_discrete_sequence=px.colors.qualitative.Vivid,
                barmode='stack',
                title='Percentage of sequences containing signals across different generated datasets',
            )

            # make bars grouped by data_origin and stacked by novelty
            # figure.update_traces(offsetgroup=df_melted["data_origin"],  # group by data_origin
            #                      opacity=1.0)

            # now reduce opacity for each novelty_memorization level
            color_map = {data_origin: px.colors.qualitative.Vivid[i]
                         for i, data_origin in enumerate(plotting_data['data_origin'].unique())}
            opacity_map = {"original": 1.0, "memorized": 1., "novel": 0.5}
            for trace in figure.data:
                data_origin, novelty = trace.name.split(' - ')
                trace.opacity = opacity_map[novelty]
                trace.marker.color = color_map[data_origin]

            figure.for_each_annotation(lambda a: a.update(text=a.text.replace("label=", "")))

            figure.update_layout(
                xaxis_title='Signal',
                yaxis_title='Frequency',
                bargap=0.2, template='plotly_white'
            )

            filename = PlotlyUtil.write_image_to_file(figure, filename)

            return ReportOutput(filename,
                                f"Summary of motif recovery for {self.name}",)

    def _get_implanted_signals(self):
        signals = []
        for signal in self.implanted_motifs_per_signal.keys():
            seeds = self.implanted_motifs_per_signal[signal]["seeds"]
            gap_sizes = self.implanted_motifs_per_signal[signal]["gap_sizes"]
            signals.append(Signal(signal, [SeedMotif(f'm{i+1}', seed=seed,
                                                     min_gap=min(gap_sizes),
                                                     max_gap=max(gap_sizes)) for i, seed in enumerate(seeds)]))
        return signals

    def check_prerequisites(self):
        location = "TrueMotifSummaryBarplot"

        run_report = True

        if 'data_origin' not in self.dataset.labels.keys():
            logging.warning(
                f"{location}: this report can only be created for a combined dataset exported with TrainGenModel "
                f"instruction. Report {self.name} will not be created.")
            run_report = False

        return run_report




import logging
import numbers
from pathlib import Path
from typing import Tuple

import pandas as pd
from sklearn.neighbors import NearestCentroid

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.ml_methods.helper_methods.FurthestNeighborClassifier import FurthestNeighborClassifier
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.clustering.ClusteringReportHandler import ClusteringReportHandler
from immuneML.workflows.instructions.clustering.ClusteringRunner import ClusteringRunner, get_features, encode_dataset
from immuneML.workflows.instructions.clustering.ClusteringState import ClusteringConfig, ClusteringState, \
    ClusteringResultPerRun, ClusteringItemResult
from immuneML.workflows.instructions.clustering.clustering_run_model import ClusteringSetting, ClusteringItem, \
    DataFrameWrapper


class ValidationHandler:
    """Handles different validation strategies for clustering."""

    def __init__(self, config: ClusteringConfig, runner: ClusteringRunner, report_handler: ClusteringReportHandler,
                 num_of_processes: int):
        self.config = config
        self.runner = runner
        self.report_handler = report_handler
        self.number_of_processes = num_of_processes

    def run_method_based_validation(self, dataset: Dataset, run_id: int, path: Path, predictions_df: pd.DataFrame,
                                    state: ClusteringState):
        """Run method-based validation."""
        print_log(f"Running method-based validation for run {run_id + 1}")
        cl_items, predictions_df = self.runner.run_all_settings(dataset, 'method_based_validation', path,
                                                                run_id, predictions_df, state)
        state.add_cl_result_per_run(run_id, 'method_based_validation',
                                    ClusteringResultPerRun(run_id, 'method_based_validation', cl_items))
        predictions_df.to_csv(state.predictions_paths[run_id]['method_based_validation'], index=False)
        print_log(f"Method-based validation for run {run_id + 1} completed.")
        return state

    def run_result_based_validation(self, dataset: Dataset, run_id: int, path: Path, predictions_df: pd.DataFrame,
                                    state: ClusteringState):
        """Run result-based validation by training a classifier on discovery clusters."""
        cl_items = {}
        analysis_desc = 'result_based_validation'
        print_log(f"Running result-based validation for run {run_id + 1}")

        for cl_setting in self.config.clustering_settings:
            # Get discovery data clustering results
            discovery_item = state.clustering_items[run_id].discovery.items[cl_setting.get_key()].item
            discovery_dataset = discovery_item.dataset

            # Train classifier on discovery data using clusters as labels
            classifier = self._train_cluster_classifier(discovery_item, cl_setting, discovery_dataset)
            cl_setting.path = PathBuilder.build(path / f"{cl_setting.get_key()}")

            # Apply classifier to validation data
            cl_item, predictions_df = self._apply_cluster_classifier(
                dataset=dataset,
                cl_setting=cl_setting,
                classifier=classifier,
                predictions_df=predictions_df,
                analysis_desc=analysis_desc,
                run_id=run_id,
                path=cl_setting.path,
                encoder=discovery_item.encoder
            )

            report_results = self.report_handler.run_item_reports(cl_item, analysis_desc, run_id, cl_setting.path,
                                                                  state)
            cl_items[cl_setting.get_key()] = ClusteringItemResult(cl_item, report_results)

        predictions_df.to_csv(state.predictions_paths[run_id][analysis_desc], index=False)
        state.add_cl_result_per_run(run_id, analysis_desc, ClusteringResultPerRun(run_id, analysis_desc, cl_items))

        print_log(f"Result-based validation for run {run_id + 1} completed.")
        return state

    def _train_cluster_classifier(self, discovery_clusters: ClusteringItem, cl_setting: ClusteringSetting,
                                  discovery_dataset: Dataset):
        """Train a classifier using discovery data clusters as labels."""
        print_log(f"Training clustering-based classifier for {cl_setting.get_key()} for result-based validation.")
        classifier = get_complementary_classifier(cl_setting)

        # Get features and cluster assignments from discovery data
        features = get_features(discovery_dataset, cl_setting)

        if len(list(set(discovery_clusters.predictions))) == 1:
            return discovery_clusters.predictions[0]
        else:
            classifier.fit(features, discovery_clusters.predictions)
            return classifier

    def _apply_cluster_classifier(self, dataset: Dataset, cl_setting: ClusteringSetting, classifier,
                                  predictions_df: pd.DataFrame, analysis_desc: str, run_id: int, path: Path,
                                  encoder: DatasetEncoder) -> Tuple[ClusteringItem, pd.DataFrame]:
        """Apply trained classifier to validation data."""

        enc_dataset = encode_dataset(dataset, cl_setting, self.number_of_processes, self.config.label_config,
                                     learn_model=False, sequence_type=self.config.sequence_type,
                                     region_type=self.config.region_type, encoder=encoder)
        features = get_features(enc_dataset, cl_setting)

        if isinstance(classifier, numbers.Number):
            predictions = [classifier] * dataset.get_example_count()
            logging.warning(f"Only one cluster found in discovery data. Assigning all validation data to "
                            f"cluster {classifier}.")
        else:
            predictions = classifier.predict(features)

        predictions_df[f'predictions_{cl_setting.get_key()}'] = predictions

        # Evaluate clustering
        performance_paths = self.runner.evaluate_clustering(predictions_df, cl_setting, features)

        cl_item = ClusteringItem(
            cl_setting=cl_setting,
            dataset=enc_dataset,
            predictions=predictions,
            external_performance=DataFrameWrapper(path=performance_paths['external']),
            internal_performance=DataFrameWrapper(path=performance_paths['internal'])
        )

        print_log(f"Result-based validation for clustering setting {cl_setting.get_key()} finished.")

        return cl_item, predictions_df


def get_complementary_classifier(cl_setting: ClusteringSetting):
    """
    Returns a complementary classifier based on the clustering method.

    Args:
        cl_setting: ClusteringSetting object containing the clustering method configuration

    Returns:
        An instance of the appropriate classifier; NearestCentroid if no matches are found
    """
    clustering_method = cl_setting.clustering_method
    method_name = clustering_method.__class__.__name__

    if method_name == 'KMeans':
        return NearestCentroid()
    elif method_name == 'AgglomerativeClustering':
        if hasattr(clustering_method.model, 'linkage'):
            if clustering_method.model.linkage == 'ward':
                return NearestCentroid()
            elif clustering_method.model.linkage == 'complete':
                if cl_setting.encoder.__class__.__name__ == 'TCRdistEncoder':
                    return FurthestNeighborClassifier(metric='precomputed')
                else:
                    return FurthestNeighborClassifier(metric='euclidean')

    return NearestCentroid()



import logging
from pathlib import Path
from typing import List

import pandas as pd
import plotly.express as px

from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.data_model.SequenceParams import RegionType
from immuneML.dsl.instruction_parsers.LabelHelper import LabelHelper
from immuneML.environment.SequenceType import SequenceType
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.KmerHelper import KmerHelper
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.SignificantFeaturesHelper import SignificantFeaturesHelper


class SignificantKmerPositions(DataReport):
    """
    Plots the number of significant k-mers (as computed by the :py:obj:`~immuneML.encodings.abundance_encoding.KmerAbundanceEncoder.KmerAbundanceEncoder` using Fisher's exact test)
    observed at each IMGT position of a given list of reference sequences.
    This report creates a stacked bar chart, where each bar represents an IMGT position, and each segment of the stack represents the observed frequency
    of one 'significant' k-mer at that position.

    **Specification arguments:**

    - reference_sequences_path (str): Path to a file containing the reference sequences,
      The file should contain one sequence per line, without a header, and without V or J genes.

    - p_values (list): The p value thresholds to be used by Fisher's exact test. Each p-value specified here will become one panel in the output figure.

    - k_values (list): Length of the k-mers (number of amino acids) created by the :py:obj:`~immuneML.encodings.abundance_encoding.KmerAbundanceEncoder.KmerAbundanceEncoder`.
      Each k-mer length will become one panel in the output figure.

    - label (dict): A label configuration. One label should be specified, and the positive_class for this label should be defined. See the YAML specification below for an example.

    - sequence_type (str): nucleotide or amino_acid

    - region_type (str): which AIRR field to consider, e.g., IMGT_CDR3 or IMGT_JUNCTION

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_significant_kmer_positions_report:
                    SignificantKmerPositions:
                        reference_sequences_path: path/to/reference/sequences.txt
                        p_values:
                            - 0.1
                            - 0.01
                            - 0.001
                            - 0.0001
                        k_values:
                            - 3
                            - 4
                            - 5
                        label: # Define a label, and the positive class for that given label
                            CMV:
                                positive_class: +
    """

    @classmethod
    def build_object(cls, **kwargs):
        location = SignificantKmerPositions.__name__

        kwargs = SignificantFeaturesHelper.parse_parameters(kwargs, location)
        kwargs = SignificantFeaturesHelper.parse_sequences_path(kwargs, "reference_sequences_path", location)
        ParameterValidator.assert_all_type_and_value(kwargs["k_values"], int, location, "k_values")

        return SignificantKmerPositions(**kwargs)

    def __init__(self, dataset: RepertoireDataset = None, reference_sequences_path: Path = None,
                 p_values: List[float] = None, k_values: List[int] = None, label: dict = None,
                 compairr_path: Path = None, result_path: Path = None, name: str = None,
                 number_of_processes: int = 1, region_type: RegionType = None, sequence_type: SequenceType = None):
        super().__init__(dataset=dataset, result_path=result_path, number_of_processes=number_of_processes, name=name)
        self.reference_sequences_path = reference_sequences_path
        self.reference_sequences = SignificantFeaturesHelper.load_sequences(reference_sequences_path)
        self.p_values = p_values
        self.k_values = k_values
        self.label = label
        self.compairr_path = compairr_path
        self.label_config = None
        self.region_type = RegionType[region_type.upper()] if isinstance(region_type, str) else region_type
        self.sequence_type = SequenceType[sequence_type.upper()] if isinstance(sequence_type, str) else sequence_type

    def check_prerequisites(self):
        if isinstance(self.dataset, RepertoireDataset):
            return True
        else:
            logging.warning(f"{SignificantKmerPositions.__name__}: report can be generated only from RepertoireDataset. "
                          f"Skipping this report...")
            return False

    def _generate(self) -> ReportResult:
        self.label_config = LabelHelper.create_label_config([self.label], self.dataset,
                                                            SignificantKmerPositions.__name__,
                                                            f"{SignificantKmerPositions.__name__}/label")

        plotting_data = self._compute_plotting_data()
        table_result = self._write_results_table(plotting_data)

        report_output_fig = self._safe_plot(plotting_data=plotting_data)
        output_figures = None if report_output_fig is None else [report_output_fig]

        return ReportResult(name=self.name,
                            info="The number of significant k-mers observed at each IMGT position of a given list "
                                 "of reference sequences.",
                            output_figures=output_figures,
                            output_tables=[table_result])

    def _compute_plotting_data(self):
        result = {"encoding": [],
                  "p-value": [],
                  "imgt_position": [],
                  "k-mer": [],
                  "count": []}

        for k in self.k_values:
            for p_value in self.p_values:
                significant_kmer_positions = self._compute_significant_kmer_positions(k, p_value)

                for imgt_pos, kmer_dict in significant_kmer_positions.items():
                    for kmer, count in kmer_dict.items():
                        result["encoding"].append(f"{k}-mer")
                        result["p-value"].append(p_value)
                        result["imgt_position"].append(str(imgt_pos))
                        result["k-mer"].append(kmer)
                        result["count"].append(count)

        return pd.DataFrame(result).astype({'imgt_position': str})

    def _get_encoder_result_path(self, k, p_value):
        result_path = self.result_path / f"{k}-mer_{p_value}"
        PathBuilder.build(result_path)
        return result_path

    def _write_results_table(self, data) -> ReportOutput:
        table_path = self.result_path / f"significant_kmer_positions_report.csv"
        data.to_csv(table_path, index=False)
        return ReportOutput(table_path,
                            "Number of significant k-mers found at each position in a set of reference sequences")

    def _plot(self, plotting_data):
        figure = px.bar(plotting_data, x="imgt_position", y="count", color="k-mer",
                        facet_row="encoding", facet_col="p-value",
                        labels={
                            "encoding": "Encoding",
                            "imgt_position": "sequence position (IMGT scheme)",
                            "count": "Number of significant k-mers observed"
                        }, template="plotly_white",
                        category_orders={
                            "imgt_position": self._get_imgt_position_order(set(plotting_data["imgt_position"]))
                        },
                        barmode="stack",
                        color_discrete_sequence=px.colors.diverging.Tealrose)

        file_path = self.result_path / f"significant_kmer_positions.html"

        file_path = PlotlyUtil.write_image_to_file(figure, file_path)

        return ReportOutput(file_path, name="Significant k-mers observed at each position in the reference sequences")

    def _get_imgt_position_order(self, imgt_positions):
        sorted_positions = sorted([float(pos) for pos in imgt_positions])
        return [str(pos_float) if int(pos_float) != pos_float else str(int(pos_float)) for pos_float in
                sorted_positions]

    def _compute_significant_kmer_positions(self, k, p_value):
        significant_kmers = self._compute_significant_kmers(k, p_value)

        results = {}

        for sequence in self.reference_sequences:
            reference_imgt_kmers = KmerHelper.create_IMGT_kmers_from_string(sequence, k,
                                                                            region_type=RegionType.IMGT_CDR3)

            for kmer, imgt_pos in reference_imgt_kmers:
                if imgt_pos not in results:
                    results[imgt_pos] = {}

                if kmer in significant_kmers:
                    if kmer in results[imgt_pos]:
                        results[imgt_pos][kmer] += 1
                    else:
                        results[imgt_pos][kmer] = 1

        return results

    def _compute_significant_kmers(self, k, p_value):
        encoder_result_path = self._get_encoder_result_path(k, p_value)
        encoder_params = SignificantFeaturesHelper._build_encoder_params(self.label_config, encoder_result_path,
                                                                         self.region_type, self.sequence_type)
        encoder = SignificantFeaturesHelper._build_kmer_encoder(self.dataset, k, p_value, encoder_params)
        sequences = pd.read_csv(encoder.relevant_sequence_path)

        return list(sequences["k-mer"])



import logging
from pathlib import Path

import numpy as np
import pandas as pd
import plotly.graph_objects as go
import yaml
from plotly.subplots import make_subplots
from sklearn.metrics import confusion_matrix

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.ml_reports.MLReport import MLReport
from immuneML.util.PathBuilder import PathBuilder


class ConfusionMatrix(MLReport):
    """
    A report that plots the confusion matrix for a trained ML method.
    Supports both binary and multiclass classification.

    **Specification arguments:**

    - alternative_label (str): optionally, the confusion matrix can be split between different values of an alternative label.
      This may be useful to compare performance across different data subsets (e.g., batches, sources).
      If specified, separate confusion matrices will be generated for each value of the alternative label. Default is None.

    **YAML specification:**

    .. code-block:: yaml

        definitions:
            reports:
                my_conf_mat_report: ConfusionMatrix

    """

    @classmethod
    def build_object(cls, **kwargs):
        return ConfusionMatrix(**kwargs)

    def __init__(self, train_dataset: Dataset = None, test_dataset: Dataset = None, method: MLMethod = None,
                 result_path: Path = None, name: str = None, hp_setting=None, label=None, number_of_processes: int = 1,
                 alternative_label: str = None):
        super().__init__(train_dataset=train_dataset, test_dataset=test_dataset, method=method,
                         result_path=result_path, name=name, hp_setting=hp_setting, label=label,
                         number_of_processes=number_of_processes)
        self.alternative_label = alternative_label

    def _generate(self):
        PathBuilder.build(self.result_path)

        y_true = np.array(self.test_dataset.encoded_data.labels[self.label.name])
        y_pred = self.method.predict(self.test_dataset.encoded_data, self.label)[self.label.name]

        labels = self.label.values
        cm = confusion_matrix(y_true, y_pred, labels=labels)

        cm_df = pd.DataFrame(cm, index=labels, columns=labels)

        heatmap_path = self._plot_confusion_matrix(cm_df)

        table_path = self.result_path / "confusion_matrix.csv"
        cm_df.rename_axis('true/predicted').reset_index().to_csv(table_path, index=False)

        output_figures = [ReportOutput(heatmap_path)]

        self._write_settings()

        if self.alternative_label is not None:
            try:
                alt_label_filename = self._generate_for_alternative_label(y_true, y_pred)
                alt_label_output = ReportOutput(alt_label_filename, f"Confusion matrix split by {self.alternative_label}")
                output_figures.append(alt_label_output)
            except Exception as e:
                logging.warning(f"Could not generate confusion matrix for alternative label "
                                f"'{self.alternative_label}': {e}")

        return ReportResult(self.name,
                            info=f"Confusion matrix for {type(self.method).__name__}",
                            output_tables=[ReportOutput(table_path, "Confusion matrix CSV")],
                            output_figures=output_figures)

    def _generate_for_alternative_label(self, y_true, y_pred):
        alt_labels = self.test_dataset.get_metadata([self.alternative_label], return_df=True)
        alt_label_values = sorted(alt_labels[self.alternative_label].unique())

        fig = make_subplots(cols=2, rows=(len(alt_label_values) + 1) // 2, subplot_titles=alt_label_values,
                            vertical_spacing=0.1,
                            shared_xaxes=True, shared_yaxes=True, x_title="Predicted Label", y_title='True Label')
        subplot_index = 0

        for alt_label_value in alt_label_values:
            indices = (alt_labels[self.alternative_label] == alt_label_value).values.astype(bool)
            y_true_subset = y_true[indices]
            y_pred_subset = y_pred[indices]

            labels = self.label.values
            cm = confusion_matrix(y_true_subset, y_pred_subset, labels=labels)

            cm_df = pd.DataFrame(cm, index=labels, columns=labels)

            fig.add_trace(go.Heatmap(z=cm_df.values, texttemplate="%{text}", text=cm_df.values, colorscale='Viridis',
                                   hovertemplate="True value: %{y}<br>Predicted value: %{x}"
                                                 "<br>Count: %{z}<extra></extra>", showscale=False,
                                   x=[str(lbl) for lbl in cm_df.index.tolist()],
                                   y=[str(lbl) for lbl in cm_df.columns.tolist()]),
                          row=(subplot_index // 2) + 1, col=(subplot_index % 2) + 1)
            subplot_index += 1

            table_path = self.result_path / f"confusion_matrix_{alt_label_value}.csv"
            cm_df.rename_axis('true/predicted').reset_index().to_csv(table_path, index=False)

        fig.update_layout(title_text=f"Confusion matrix across {self.alternative_label} values",
                          template="plotly_white")
        filename = self.result_path / f"confusion_matrix_{self.alternative_label}.html"
        PlotlyUtil.write_image_to_file(fig, filename, self.test_dataset.get_example_count())
        return filename

    def _plot_confusion_matrix(self, cm_df: pd.DataFrame):

        fig = go.Figure(go.Heatmap(z=cm_df.values, texttemplate="%{text}", text=cm_df.values, colorscale='Viridis',
                                   hovertemplate="True value: %{y}<br>Predicted value: %{x}"
                                                 "<br>Count: %{z}<extra></extra>", showscale=False,
                                   x=[str(lbl) for lbl in cm_df.index.tolist()],
                                   y=[str(lbl) for lbl in cm_df.columns.tolist()]))

        fig.update_layout(title_text="Confusion Matrix", xaxis_title="Predicted class", yaxis_title="True class",
                          template="plotly_white")

        filename = self.result_path / "confusion_matrix.html"
        filename = PlotlyUtil.write_image_to_file(fig, filename)

        return filename

    def _write_settings(self):
        if self.hp_setting is not None:
            file_path = self.result_path / "settings.yaml"
            with file_path.open("w") as file:
                yaml.dump({"preprocessing": self.hp_setting.preproc_sequence_name,
                           "encoder": self.hp_setting.encoder_name,
                           "ml_method": self.hp_setting.ml_method_name},
                          file)

    def check_prerequisites(self):
        if self.test_dataset is None or self.label is None:
            logging.warning("ConfusionMatrixReport requires a test dataset and a specified label.")
            return False
        return True



# quality: gold

import abc
import logging
import uuid
from dataclasses import fields
from multiprocessing import Pool
from pathlib import Path
from typing import Type

import pandas as pd
from bionumpy import AminoAcidEncoding

from immuneML.IO.dataset_import.DatasetImportParams import DatasetImportParams
from immuneML.data_model.AIRRSequenceSet import AIRRSequenceSet, AminoAcidXEncoding, DNANEncoding
from immuneML.data_model.SequenceSet import Repertoire, build_dynamic_airr_sequence_set_dataclass
from immuneML.data_model.bnp_util import bnp_write_to_file, read_yaml, write_dataset_yaml
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.ElementDataset import SequenceDataset, ReceptorDataset, ElementDataset
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.util.ImportHelper import ImportHelper
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class DataImport(metaclass=abc.ABCMeta):

    def __init__(self, params: dict, dataset_name: str):
        self.params = DatasetImportParams.build_object(**params) if isinstance(params, dict) else params
        self.dataset_name = dataset_name

    def import_dataset(self) -> Dataset:

        if self.params.dataset_file is not None:
            assert self.params.dataset_file.is_file(), \
                f"DataImport: dataset_file was specified but not found: {self.params.dataset_file}"
            dataset = self.import_dataset_from_yaml()
        else:
            if self.params.is_repertoire is None and self.params.metadata_file is not None:
                self.params.is_repertoire = True

            if self.params.is_repertoire:
                dataset = self.import_repertoire_dataset()
            elif self.params.paired:
                dataset = self.import_receptor_dataset()
            else:
                dataset = self.import_sequence_dataset()

        dataset.labels['organism'] = self.params.organism

        return dataset

    def import_dataset_from_yaml(self):
        dataset_metadata = read_yaml(self.params.dataset_file)
        if dataset_metadata["dataset_type"] == "RepertoireDataset":
            dataset = self.import_repertoire_dataset()
        elif dataset_metadata["dataset_type"] == "ReceptorDataset":
            dataset = self.import_receptor_dataset()
        elif dataset_metadata["dataset_type"] == "SequenceDataset":
            dataset = self.import_sequence_dataset()
        else:
            raise TypeError(
                f"DataImport: Dataset type '{dataset_metadata['dataset_type']}' not recognized (expected one of: RepertoireDataset, ReceptorDataset, SequenceDataset)")

        return dataset

    def import_repertoire_dataset(self) -> RepertoireDataset:
        if self.params.dataset_file is not None and self.params.dataset_file.is_file():
            imported_dataset_yaml = read_yaml(self.params.dataset_file)

            self.params.path = self.params.dataset_file.parent

            if self.params.metadata_file is None:
                self.params.metadata_file = self.params.dataset_file.parent / imported_dataset_yaml['metadata_file']

            labels = imported_dataset_yaml['labels']
            metadata = self.import_repertoire_metadata(self.params.metadata_file)
            repertoires = []
            for index, row in metadata.iterrows():
                filename = ImportHelper.get_repertoire_filename_from_metadata_row(row, self.params)
                metadata_filename = filename.with_suffix('.yaml')
                metadata_from_file = read_yaml(metadata_filename) if metadata_filename.is_file() else {}
                rep_metadata = {**metadata_from_file, **row.to_dict()}
                repertoires.append(Repertoire(data_filename=filename, metadata_filename=metadata_filename,
                                              metadata=rep_metadata, identifier=row['identifier']))

                if repertoires[-1].get_element_count() == 0:
                    logging.warning(f"Repertoire {repertoires[-1].data_filename} contains 0 sequences. It is "
                                    f"recommended to remove this repertoire from the dataset.")

            new_metadata_file = self.params.metadata_file
            dataset_filename = self.params.dataset_file
            dataset_yaml = imported_dataset_yaml

        else:
            metadata = self.import_repertoire_metadata(self.params.metadata_file)
            repertoires = self.load_repertoires(metadata)
            new_metadata_file = ImportHelper.make_new_metadata_file(repertoires, metadata, self.params.result_path,
                                                                    self.dataset_name)
            # type_dict = self.determine_repertoire_type_dict(repertoires)
            labels = self.determine_repertoire_dataset_labels(metadata)
            identifier = None

            dataset_filename = PathBuilder.build(self.params.result_path) / f"{self.dataset_name}.yaml"
            dataset_yaml = RepertoireDataset.create_metadata_dict(metadata_file=new_metadata_file,
                                                                  labels=labels,
                                                                  name=self.dataset_name,
                                                                  identifier=identifier)
            write_dataset_yaml(dataset_filename, dataset_yaml)

        return RepertoireDataset(labels=labels,
                                 repertoires=repertoires, metadata_file=new_metadata_file, name=self.dataset_name,
                                 dataset_file=dataset_filename, identifier=dataset_yaml["identifier"])

    def import_repertoire_metadata(self, metadata_file_path):
        try:
            metadata = pd.read_csv(metadata_file_path, sep=",")
            if "identifier" in metadata.columns:
                assert metadata['identifier'].is_unique, \
                    (f"DataImport: if the field 'identifier' is supplied, each repertoire must have "
                     f"a unique identifier (found {len(set(list(metadata['identifier'])))} unique "
                     f"identifiers for {len(list(metadata['identifier']))} repertoires).")
        except Exception as e:
            raise Exception(f"{e}\nAn error occurred while reading in the metadata file {self.params.metadata_file}. "
                            f"Please see the error log above for more details on this error and the documentation "
                            f"for the expected format of the metadata.").with_traceback(e.__traceback__)

        ParameterValidator.assert_keys_present(metadata.columns.tolist(), ["filename"], self.__class__.__name__,
                                               f'{self.dataset_name}: params: metadata_file')
        return metadata

    def load_repertoires(self, metadata):
        PathBuilder.build(self.params.result_path / "repertoires/")

        with Pool(self.params.number_of_processes) as pool:
            repertoires = pool.map(self.load_repertoire_object, [row for _, row in metadata.iterrows()])

        return repertoires

    def determine_repertoire_type_dict(self, repertoires):
        try:
            if all(repertoires[0].metadata['type_dict_dynamic_fields'] == rep.metadata['type_dict_dynamic_fields'] for
                   rep
                   in repertoires[1:]):
                return repertoires[0].metadata['type_dict_dynamic_fields']
            else:
                raise RuntimeError()
        except Exception as e:
            logging.warning(f'{DataImport.__name__}: dynamic fields for the dataset {self.dataset_name} could not be '
                            f'extracted, some repertoires have different fields.')
            return {}

    def determine_repertoire_dataset_labels(self, metadata, imported_labels=None):
        potential_label_names = list(
            set(metadata.columns.tolist()) - {"filename", "type_dict_dynamic_fields", "identifier", "subject_id"})
        potential_labels = {key: list(set(metadata[key].values.tolist())) for key in potential_label_names}

        if imported_labels is not None:
            labels = imported_labels
            if any(label not in potential_label_names for label in imported_labels):
                logging.warning(f"{DataImport.__name__}: an error occurred when importing dataset {self.dataset_name}. "
                                f"Labels specified in the dataset file ({imported_labels}) could not be found in the repertoire "
                                f"fields. Proceeding with the following labels: {potential_labels}.")
                labels = potential_labels
        else:
            labels = potential_labels

        return labels

    def import_element_dataset(self, dataset_class: Type, filter_func=None):
        if self.params.dataset_file is not None and self.params.dataset_file.is_file():
            filenames = [Path(read_yaml(self.params.dataset_file)["filename"])]
            if self.params.path.is_file():
                filenames[0] = self.params.path
            elif self.params.path:
                filenames[0] = self.params.path / Path(filenames[0]).name
            else:
                filenames[0] = self.params.dataset_file.parent / Path(filenames[0]).name
            assert filenames[0].is_file(), \
                f"DataImport: filename {filenames[0]} specified in dataset file was not found."
        else:
            filenames = ImportHelper.get_sequence_filenames(self.params.path, self.dataset_name)

        final_data_dict = self._construct_element_dataset_data_dict(filenames, filter_func)

        bnp_dc, type_dict = build_dynamic_airr_sequence_set_dataclass(final_data_dict)
        filename = self._write_element_dataset_tsv_file(final_data_dict, bnp_dc)

        possible_labels = {key: list(set(final_data_dict[key])) for key in type_dict.keys()}
        logging.info(f"All possible labels for dataset '{self.dataset_name}' are: {list(possible_labels.keys())}")

        if self.params.label_columns:
            label_variants = self.params.label_columns + [ImportHelper.get_standardized_name(label_name) for label_name
                                                          in self.params.label_columns]
            possible_labels = {key: value for key, value in possible_labels.items() if key in label_variants}

        dataset_filename = self._write_element_dataset_metadata_file(dataset_class, filename, type_dict,
                                                                     possible_labels)
        metadata = read_yaml(dataset_filename)

        logging.info(f"Displayed labels for dataset '{self.dataset_name}' are: {list(metadata['labels'].keys())}")

        return dataset_class(name=self.dataset_name, bnp_dataclass=bnp_dc, dataset_file=dataset_filename,
                             dynamic_fields=type_dict, filename=filename, labels=metadata['labels'])

    def import_sequence_dataset(self) -> SequenceDataset:
        return self.import_element_dataset(SequenceDataset)

    def import_receptor_dataset(self) -> ReceptorDataset:
        return self.import_element_dataset(ReceptorDataset,
                                           lambda df: ImportHelper.filter_illegal_receptors(df, self.params.receptor_chains))

    def _construct_element_dataset_data_dict(self, filenames, filter_func) -> dict:
        final_df = None

        for filename in filenames:
            df = self.load_sequence_dataframe(filename)
            if filter_func:
                df = filter_func(df)
            final_df = pd.concat([final_df, df])

        return final_df.to_dict(orient='list')

    def _write_element_dataset_tsv_file(self, final_data_dict, bnp_dc) -> Path:
        PathBuilder.build(self.params.result_path)

        data = bnp_dc(**final_data_dict)
        data_filename = self.params.result_path / f'{self.dataset_name}.tsv'
        bnp_write_to_file(data_filename, data)

        return data_filename

    def _write_element_dataset_metadata_file(self, dataset_class, filename, type_dict, possible_labels):
        dataset_filename = self.params.result_path / f"{self.dataset_name}.yaml"

        if self.params.dataset_file:
            metadata = read_yaml(self.params.dataset_file)
        else:
            metadata = ElementDataset.create_metadata_dict(dataset_class=dataset_class.__name__,
                                                           filename=filename,
                                                           type_dict=type_dict,
                                                           name=self.dataset_name,
                                                           labels=possible_labels)
        write_dataset_yaml(dataset_filename, metadata)
        return dataset_filename

    def load_repertoire_object(self, metadata_row: pd.Series) -> Repertoire:
        try:
            filename = ImportHelper.get_repertoire_filename_from_metadata_row(metadata_row, self.params)

            dataframe = self.load_sequence_dataframe(filename)

            repertoire_inputs = {**{"metadata": metadata_row.to_dict(),
                                    "path": self.params.result_path / "repertoires/",
                                    "filename_base": filename.stem}, **dataframe.to_dict(orient='list')}

            if "identifier" in metadata_row:
                repertoire_inputs["identifier"] = metadata_row["identifier"]

            repertoire = Repertoire.build(**repertoire_inputs)

            if repertoire.get_element_count() == 0:
                logging.warning(
                    f"Repertoire {repertoire.data_filename} contains 0 sequences. It is recommended to remove this "
                    f"repertoire from the dataset. ")

            return repertoire
        except Exception as e:
            raise RuntimeError(
                f"{self.__class__.__name__}: error when importing file {Path(metadata_row['filename']).absolute()}: "
                f"\n{e}.\n")

    def load_sequence_dataframe(self, filename: Path):

        df = self.load_file(filename)
        df = self.preprocess_file(df)

        df = ImportHelper.standardize_column_names(df)
        df = ImportHelper.add_cdr3_from_junction(df)
        df = ImportHelper.drop_empty_sequences(df, self.params.import_empty_aa_sequences,
                                               self.params.import_empty_nt_sequences,
                                               self.params.region_type)
        df = ImportHelper.drop_illegal_character_sequences(df, self.params.import_illegal_characters,
                                                           self.params.import_with_stop_codon, self.params.region_type)
        df = ImportHelper.filter_illegal_sequences(df, self.params, self.dataset_name)
        df = ImportHelper.extract_locus_from_data(df, self.params, self.dataset_name)
        df = ImportHelper.add_default_fields_for_airr_seq_set(df)

        assert df.columns.shape[0] == df.columns.unique().shape[0], \
            f"There are non-unique columns in the imported file {filename.name}: {df.columns.tolist()}"

        return df

    def preprocess_file(self, df: pd.DataFrame) -> pd.DataFrame:
        if 'sequence_id' in df.columns:
            missing_seq_ids = df['sequence_id'].isnull() | df['sequence_id'].eq('')
            df.loc[missing_seq_ids, 'sequence_id'] = [uuid.uuid4().hex for _ in range(missing_seq_ids.sum())]
        return df

    def load_file(self, filename: Path) -> pd.DataFrame:
        try:
            df = pd.read_csv(str(filename), sep=self.params.separator, iterator=False,
                             usecols=self.params.columns_to_load, engine='python')
        except ValueError:

            df = pd.read_csv(str(filename), sep=self.params.separator, iterator=False, engine='python')

            expected = [e for e in self.params.columns_to_load if e not in list(df.columns)]
            df = df[[col for col in self.params.columns_to_load if col in df.columns]]

            logging.warning(
                f"{self.__class__.__name__}: expected to find the following column(s) in the input file "
                f"'{filename.name}', which were not found: {expected}. The following columns were imported instead: "
                f"{list(df.columns)}. \nTo remove this warning, add the relevant columns to the input file, "
                f"or change which columns are imported under 'datasets/{self.dataset_name}/params/columns_to_load' and "
                f"'datasets/{self.dataset_name}/params/column_mapping'.")

        if hasattr(self.params, "column_mapping") and self.params.column_mapping is not None:
            df.rename(columns=self.params.column_mapping, inplace=True)

        df = df.dropna(how='all')  # remove all fully empty rows

        df = ImportHelper.standardize_bool_values(df)
        df = ImportHelper.standardize_none_values(df)
        df = ImportHelper.standardize_string_values(df, filename.name)

        df = self._convert_types(df)

        return df

    def _convert_types(self, df: pd.DataFrame) -> pd.DataFrame:
        str_cols = [f.name for f in fields(AIRRSequenceSet)
                    if f.type in [str, AminoAcidEncoding, AminoAcidXEncoding, DNANEncoding] and f.name in df.columns]

        df = df.astype({col: str for col in str_cols})
        df.loc[:, str_cols] = df.loc[:, str_cols].replace('nan', '').replace('-1.0', '')

        encoded_cols = [f for f, t in AIRRSequenceSet.get_field_type_dict().items()
                        if t in [AminoAcidXEncoding, AminoAcidEncoding, DNANEncoding] and f in df.columns]

        df.loc[:, encoded_cols] = df.loc[:, encoded_cols].apply(lambda x: x.str.upper())

        invalid_cols = df.columns[~df.map(type).nunique().eq(1)]
        df[invalid_cols] = df[invalid_cols].astype(str)

        int_cols = [f.name for f in fields(AIRRSequenceSet) if f.type == int and f.name in df.columns]
        df[int_cols] = df[int_cols].astype(int)

        return df



import copy
import glob
import shutil
import subprocess
from pathlib import Path
from typing import List

import pandas as pd

from immuneML.IO.dataset_export.AIRRExporter import AIRRExporter
from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.SequenceSet import ReceptorSequence, Repertoire
from immuneML.data_model.datasets.ElementDataset import SequenceDataset
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.encodings.reference_encoding.MatchedReferenceUtil import MatchedReferenceUtil
from immuneML.preprocessing.Preprocessor import Preprocessor
from immuneML.util.CompAIRRHelper import CompAIRRHelper
from immuneML.util.CompAIRRParams import CompAIRRParams
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class ReferenceSequenceAnnotator(Preprocessor):
    """
    Annotates each sequence in each repertoire if it matches any of the reference sequences provided as input parameter. This report uses CompAIRR internally. To match CDR3 sequences (and not JUNCTION), CompAIRR v1.10 or later is needed.

    **Specification arguments:**

    - reference_sequences (dict): A dictionary describing the reference dataset file. Import should be specified the same way as regular dataset import. It is only allowed to import a receptor dataset here (i.e., is_repertoire is False and paired is True by default, and these are not allowed to be changed).

    - max_edit_distance (int): The maximum edit distance between a target sequence (from the repertoire) and the reference sequence.

    - compairr_path (str): optional path to the CompAIRR executable. If not given, it is assumed that CompAIRR has been installed such that it can be called directly on the command line with the command 'compairr', or that it is located at /usr/local/bin/compairr.

    - threads (int): how many threads to be used by CompAIRR for sequence matching

    - ignore_genes (bool): Whether to ignore V and J gene information. If False, the V and J genes between two receptor chains have to match. If True, gene information is ignored. By default, ignore_genes is False.

    - output_column_name (str): in case there are multiple annotations, it is possible here to define the name of the column in the output repertoire files for this specific annotation

    - repertoire_batch_size (int): how many repertoires to process simultaneously; depending on the repertoire size, this parameter might be use to limit the memory usage

    - region_type (str): which region type to check, default is IMGT_CDR3


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        preprocessing_sequences:
            my_preprocessing:
                - step1:
                    ReferenceSequenceAnnotator:
                        reference_sequences:
                            format: VDJDB
                            params:
                                path: path/to/file.csv
                        compairr_path: optional/path/to/compairr
                        ignore_genes: False
                        max_edit_distance: 0
                        output_column_name: matched
                        threads: 4
                        repertoire_batch_size: 5
                        region_type: IMGT_CDR3

    """

    def __init__(self, reference_sequences: List[ReceptorSequence], max_edit_distance: int, compairr_path: str, ignore_genes: bool, threads: int,
                 output_column_name: str, repertoire_batch_size: int, region_type: RegionType):
        super().__init__()
        self._reference_sequences = reference_sequences
        self._max_edit_distance = max_edit_distance
        self._output_column_name = output_column_name
        self._repertoire_batch_size = repertoire_batch_size
        self._region_type = region_type
        self._compairr_params = CompAIRRParams(compairr_path=CompAIRRHelper.determine_compairr_path(compairr_path), keep_compairr_input=True,
                                               differences=max_edit_distance, indels=False, ignore_counts=True, ignore_genes=ignore_genes,
                                               threads=threads, output_filename="compairr_out.tsv", log_filename="compairr_log.txt",
                                               output_pairs=False, do_repertoire_overlap=False, do_sequence_matching=True, pairs_filename="pairs.tsv")

    @classmethod
    def build_object(cls, **kwargs):
        ParameterValidator.assert_keys(list(kwargs.keys()),
                                       ['reference_sequences', 'max_edit_distance', 'compairr_path', 'ignore_genes', 'output_column_name', 'threads',
                                        'repertoire_batch_size', 'region_type'],
                                       ReferenceSequenceAnnotator.__name__, ReferenceSequenceAnnotator.__name__)
        ParameterValidator.assert_in_valid_list(kwargs['region_type'].upper(), [el.name for el in RegionType],
                                                ReferenceSequenceAnnotator.__name__, 'region_type')
        ref_seqs = MatchedReferenceUtil.prepare_reference(reference_params=kwargs['reference_sequences'],
                                                          location=ReferenceSequenceAnnotator.__name__, paired=False)
        return ReferenceSequenceAnnotator(**{**kwargs, 'reference_sequences': ref_seqs,
                                             'region_type': RegionType[kwargs['region_type'].upper()]})

    def process_dataset(self, dataset: RepertoireDataset, result_path: Path, number_of_processes=1) -> RepertoireDataset:
        self._compairr_params.is_cdr3 = self._region_type == RegionType.IMGT_CDR3
        sequence_filepath = self._prepare_sequences_for_compairr(result_path / 'tmp')
        repertoires_filepaths = self._prepare_repertoires_for_compairr(dataset, result_path / 'tmp')

        compairr_output_files = self._annotate_repertoires(sequence_filepath, repertoires_filepaths, result_path)

        processed_dataset = self._make_annotated_dataset(dataset, result_path, compairr_output_files)

        shutil.rmtree(result_path / 'tmp')

        return processed_dataset

    def _annotate_repertoires(self, sequences_filepath, repertoire_filepaths: List[Path], result_path: Path):
        tmp_path = PathBuilder.build(result_path / 'tmp')
        updated_output_files = []

        for index, rep_file in enumerate(repertoire_filepaths):
            batch_tmp_path = PathBuilder.build(tmp_path / str(index))
            args = CompAIRRHelper.get_cmd_args(self._compairr_params, [rep_file, sequences_filepath], batch_tmp_path)
            compairr_result = subprocess.run(args, capture_output=True, text=True)
            output_file = CompAIRRHelper.verify_compairr_output_path(compairr_result, self._compairr_params, batch_tmp_path)
            updated_output_file = result_path / f'updated_compairr_output_{index}.tsv'

            with open(output_file, 'r') as file:
                output_lines = file.readlines()

            with updated_output_file.open(mode='w') as file:
                output_lines[0] = output_lines[0].replace("#", '')
                file.writelines(output_lines)

            updated_output_files.append(updated_output_file)

        return updated_output_files

    def _make_annotated_dataset(self, dataset: RepertoireDataset, result_path: Path, compairr_output_files: List[Path]) -> RepertoireDataset:
        repertoires = []
        repertoire_path = PathBuilder.build(result_path / 'repertoires')

        for index, repertoire in enumerate(dataset.repertoires):

            compairr_out_df = pd.read_csv(compairr_output_files[index], sep='\t', comment="#")

            data = repertoire.data.add_fields({self._output_column_name: compairr_out_df.iloc[:, 1].values.astype(int)},
                                              {self._output_column_name: int})
            type_dict = type(data).get_field_type_dict(all_fields=False)

            repertoires.append(Repertoire.build_from_dc_object(repertoire_path, metadata=repertoire.metadata,
                                                               filename_base=repertoire.data_filename.name, data=data,
                                                               type_dict=type_dict))

        return RepertoireDataset.build_from_objects(**{'repertoires': repertoires, 'path': result_path})

    def _prepare_sequences_for_compairr(self, result_path: Path) -> Path:
        path = PathBuilder.build(result_path) / 'reference_sequences.tsv'

        # TODO: remove this when import is refactored, this now ensures that string matching is done on sequences as imported
        reference_sequences = []
        for seq in self._reference_sequences:
            tmp_seq = copy.deepcopy(seq)
            tmp_seq.duplicate_count = seq.duplicate_count if not self._compairr_params.ignore_counts else 1
            reference_sequences.append(tmp_seq)

        SequenceDataset.build_from_objects(reference_sequences, PathBuilder.build(result_path),
                                           region_type=self._region_type)

        result_files = glob.glob(str(result_path / "*.tsv"))
        assert len(result_files) == 1, \
            f"Error occurred while exporting sequences for matching using CompAIRR. Resulting files: {result_files}"
        shutil.move(result_files[0], path)

        return path

    def _prepare_repertoires_for_compairr(self, dataset: RepertoireDataset, result_path: Path) -> List[Path]:
        PathBuilder.build(result_path)
        paths = []
        for i, repertoire in enumerate(dataset.repertoires):
            path = result_path / f'repertoires_{i}.tsv'
            CompAIRRHelper.write_repertoire_file(repertoires=[repertoire], filename=path,
                                                 compairr_params=self._compairr_params, export_sequence_id=True)
            paths.append(path)
        return paths

    def _check_column_name(self, dataset):
        for repertoire in dataset.repertoires:
            assert repertoire.get_attribute(self._output_column_name) is None, \
                (f"{ReferenceSequenceAnnotator.__name__}: attribute {self._output_column_name} already exists in "
                 f"repertoire ({repertoire.identifier}); choose another name.")



import copy
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict

from immuneML.IO.dataset_export.AIRRExporter import AIRRExporter
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.environment.SequenceType import SequenceType
from immuneML.ml_methods.generative_models import GenerativeModel
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.reports.ml_reports.MLReport import MLReport
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.Instruction import Instruction


@dataclass
class ApplyGenModelState:
    result_path: Path
    name: str
    gen_examples_count: int
    model_path: Path = None
    generated_dataset: Dataset = None
    exported_datasets: Dict[str, Path] = field(default_factory=dict)
    report_results: dict = field(default_factory=lambda: {'data_reports': [], 'ml_reports': []})


class ApplyGenModelInstruction(Instruction):
    """

    ApplyGenModel instruction implements applying generative AIRR models on the sequence level.

    This instruction takes as input a trained model (trained in the :ref:`TrainGenModel` instruction)
    which will be used for generating data and the number of sequences to be generated.
    It can also produce reports of the applied model and reports of generated sequences.


    **Specification arguments:**

    - gen_examples_count (int): how many examples (sequences, repertoires) to generate from the applied model

    - reports (list): list of report ids (defined under definitions/reports) to apply after generating
      gen_examples_count examples; these can be data reports (to be run on generated examples), ML reports (to be run
      on the fitted model)

    - ml_config_path (str): path to the trained model in zip format (as provided by TrainGenModel instruction)

    **YAML specification:**

    .. highlight:: yaml
    .. code-block:: yaml

        instructions:
            my_apply_gen_model_inst: # user-defined instruction name
                type: ApplyGenModel
                gen_examples_count: 100
                ml_config_path: ./config.zip
                reports: [data_rep1, ml_rep2]

    """

    def __init__(self, method: GenerativeModel = None, reports: list = None, result_path: Path = None,
                 name: str = None, gen_examples_count: int = None):
        self.state = ApplyGenModelState(result_path, name, gen_examples_count)
        self.method = method
        self.reports = reports
        self.generated_dataset = None

    def run(self, result_path: Path) -> ApplyGenModelState:
        self._set_path(result_path)
        self._gen_data()
        self._export_generated_dataset()
        self._run_reports()

        return self.state

    def _gen_data(self):
        dataset = self.method.generate_sequences(self.state.gen_examples_count, 1,
                                                 self.state.result_path / 'generated_sequences',
                                                 SequenceType.AMINO_ACID, False)

        self.generated_dataset = dataset
        print_log(f"{self.state.name}: generated {self.state.gen_examples_count} examples from the fitted model",
                  True)

    def _export_generated_dataset(self):
        AIRRExporter.export(self.state.generated_dataset, self.state.result_path / f'exported_gen_dataset')
        self.state.exported_datasets['generated_dataset'] = self.state.result_path / 'exported_gen_dataset'

    def _run_reports(self):
        report_path = self._get_reports_path()
        for report in self.reports:
            report.result_path = report_path
            if isinstance(report, DataReport):
                rep = copy.deepcopy(report)
                rep.dataset = self.generated_dataset
                rep.name = rep.name + " (generated dataset)"
                self.state.report_results['data_reports'].append(rep.generate_report())
            elif isinstance(report, MLReport):
                rep = copy.deepcopy(report)
                rep.method = self.method
                rep.name = rep.name
                self.state.report_results['ml_reports'].append(rep.generate_report())

        self._print_report_summary_log()

    def _print_report_summary_log(self):
        if len(self.reports) > 0:
            gen_rep_count = len(self.state.report_results['ml_reports']) + len(
                self.state.report_results['data_reports'])
            print_log(f"{self.state.name}: generated {gen_rep_count} reports.", True)

    def _get_reports_path(self) -> Path:
        return PathBuilder.build(self.state.result_path / 'reports')

    def _set_path(self, result_path):
        self.state.result_path = PathBuilder.build(result_path / self.state.name)



import random
import typing
from dataclasses import dataclass
from typing import List, Union

from immuneML import Constants
from immuneML.environment.SequenceType import SequenceType
from immuneML.simulation.implants.Motif import Motif
from immuneML.simulation.implants.MotifInstance import MotifInstanceGroup


@dataclass
class Signal:
    """
    A signal represents a collection of motifs, and optionally, position weights showing where one
    of the motifs of the signal can occur in a sequence.
    The signals are defined under :code:`definitions/signals`.

    A signal is associated with a metadata label, which is assigned to a receptor or repertoire.
    For example antigen-specific/disease-associated (receptor) or diseased (repertoire).

    .. note:: IMGT positions

        To use sequence position weights, IMGT positions should be explicitly specified as strings, under quotation marks, to allow for all positions to be properly distinguished.

    **Specification arguments:**

    - motifs (list): A list of the motifs associated with this signal, either defined by seed or by position weight matrix. Alternatively, it can be a list of a list of motifs, in which case the motifs in the same sublist (max 2 motifs) have to co-occur in the same sequence

    - sequence_position_weights (dict): a dictionary specifying for each IMGT position in the sequence how likely it is for the signal to be there. If the position is not present in the sequence, the probability of the signal occurring at that position will be redistributed to other positions with probabilities that are not explicitly set to 0 by the user.

    - v_call (str): V gene with allele if available that has to co-occur with one of the motifs for the signal to exist; can be used in combination with rejection sampling, or full sequence implanting, otherwise ignored; to match in a sequence for rejection sampling, it is checked if this value is contained in the same field of generated sequence;

    - j_call (str): J gene with allele if available that has to co-occur with one of the motifs for the signal to exist; can be used in combination with rejection sampling, or full sequence implanting, otherwise ignored; to match in a sequence for rejection sampling, it is checked if this value is contained in the same field of generated sequence;

    - source_file (str): path to the file where the custom signal function is; cannot be combined with the arguments listed above (motifs, v_call, j_call, sequence_position_weights)

    - is_present_func (str): name of the function from the source_file file that will be used to specify the signal; the function's signature must be:

    .. code-block:: python

        def is_present(sequence_aa: str, sequence: str, v_call: str, j_call: str) -> bool:
            # custom implementation where all or some of these arguments can be used

    - clonal_frequency (dict): clonal frequency in Ligo is simulated through `scipy's zeta distribution function for generating random numbers <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.zipf.html>`_, with parameters provided under clonal_frequency parameter. If clonal frequency should not be used, this parameter can be None

    .. code-block:: yaml

      clonal_frequency:
        a: 2 # shape parameter of the distribution
        loc: 0 # 0 by default but can be used to shift the distribution


    **YAML specification:**

    .. code-block:: yaml

        definitions:
            signals:
                my_signal:
                    motifs:
                        - my_simple_motif
                        - my_gapped_motif
                    sequence_position_weights:
                        '109': 0.5
                        '110': 0.5
                    v_call: TRBV1
                    j_call: TRBJ1
                    clonal_frequency:
                        a: 2
                        loc: 0
                signal_with_custom_func:
                    source_file: signal_func.py
                    is_present_func: is_signal_present
                    clonal_frequency:
                        a: 2
                        loc: 0

    """
    id: str
    motifs: List[Union[Motif, List[Motif]]] = None
    sequence_position_weights: dict = None
    v_call: str = None
    j_call: str = None
    clonal_frequency: dict = None
    is_present_custom_func: typing.Callable = None

    def get_all_motif_instances(self, sequence_type: SequenceType):
        motif_instances = []
        for motif_group in self.motifs:
            if isinstance(motif_group, list):
                motif_instances.append(MotifInstanceGroup([motif.get_all_possible_instances(sequence_type) for motif in motif_group]))
            else:
                motif_instances.append(motif_group.get_all_possible_instances(sequence_type))
        return motif_instances

    def make_motif_instances(self, count, sequence_type: SequenceType):
        return [motif.instantiate_motif(sequence_type=sequence_type) for motif in random.choices(self.motifs, k=count)]

    def __hash__(self):
        return hash(self.id)

    def __str__(self):
        return str(vars(self))


@dataclass
class SignalPair:
    signal1: Signal
    signal2: Signal

    @property
    def id(self) -> str:
        return Constants.SIGNAL_DELIMITER.join(sorted([self.signal1.id, self.signal2.id]))

    @property
    def v_call(self):
        return [el for el in sorted(list({self.signal1.v_call, self.signal2.v_call})) if el is not None] \
            if self.signal1.v_call is not None and self.signal2.v_call is not None else None

    @property
    def j_call(self):
        return [el for el in sorted(list({self.signal1.j_call, self.signal2.j_call})) if el is not None] \
            if self.signal1.j_call is not None and self.signal2.j_call is not None else None

    @property
    def clonal_frequency(self):
        return random.choice([self.signal1, self.signal2]).clonal_frequency

    def __hash__(self):
        return hash(tuple(sorted([self.signal1.id, self.signal2.id])))



from dataclasses import dataclass
from typing import List, Union

from immuneML.environment.SequenceType import SequenceType
from immuneML.simulation.SimConfigItem import SimConfigItem
from immuneML.simulation.simulation_strategy.SimulationStrategy import SimulationStrategy


@dataclass
class SimConfig:
    """
    The simulation config defines all parameters of the simulation.
    It can contain one or more simulation config items, which define groups of repertoires or receptors
    that have the same simulation parameters, such as signals, generative model, clonal frequencies, and noise parameters.


    **Specification arguments:**

    - sim_items (dict): a list of SimConfigItems defining individual units of simulation

    - is_repertoire (bool): whether the simulation is on a repertoire (person) or sequence/receptor level

    - paired: if the simulation should output paired data, this parameter should contain a list of a list of sim_item pairs referenced by name that should be combined; if paired data is not needed, then it should be False

    - sequence_type (str): either amino_acid or nucleotide

    - simulation_strategy (str): either RejectionSampling or Implanting, see the tutorials for more information on choosing one of these

    - keep_p_gen_dist (bool): if possible, whether to keep the distribution of generation probabilities of the sequences the same as provided by the model without any signals

    - p_gen_bin_count (int): if keep_p_gen_dist is true, how many bins to use to approximate the generation probability distribution

    - remove_seqs_with_signals (bool): if true, it explicitly controls the proportions of signals in sequences and removes any accidental occurrences

    - species (str): species that the sequences come from; used to select correct genes to export full length sequences; default is 'human'

    - implanting_scaling_factor (int): determines in how many receptors to implant the signal in reach iteration; this is computed as number_of_receptors_needed_for_signal * implanting_scaling_factor; useful when using Implanting simulation strategy in combination with importance sampling, since the generation probability of some receptors with implanted signals might be very rare and those receptors might end up not being kept often with importance sampling; this parameter is only used when keep_p_gen_dist is set to True


    **YAML specification:**

    .. indent-with-spaces
    .. code-block:: yaml

        definitions:
            simulations:
                sim1:
                    is_repertoire: false
                    paired: false
                    sequence_type: amino_acid
                    simulation_strategy: RejectionSampling
                    sim_items:
                        sim_item1: # group of sequences with same simulation params
                            generative_model:
                                chain: beta
                                default_model_name: humanTRB
                                model_path: null
                                type: OLGA
                            number_of_examples: 100
                            seed: 1002
                            signals:
                                signal1: 1

    """
    sim_items: List[SimConfigItem] = None
    identifier: str = None
    is_repertoire: bool = None
    paired: Union[bool, List[List[str]]] = None
    sequence_type: SequenceType = None
    simulation_strategy: SimulationStrategy = None
    p_gen_bin_count: int = None
    keep_p_gen_dist: bool = None
    remove_seqs_with_signals: bool = None
    species: str = None
    implanting_scaling_factor: int = None

    def __str__(self):
        return ",\n".join(str(simulation_item) for simulation_item in self.sim_items)

    def get_total_seq_count_for_signal(self, signal_id: str, model_name: str) -> int:
        sim_item_names = model_name.split("__")
        total_count = 0
        for sim_item_name in sim_item_names:
            total_count += self._get_seq_count_for_sim_item(signal_id, sim_item_name)
        return round(total_count)

    def get_total_seq_count(self, model_name: str) -> int:
        sim_item_names = model_name.split("__")
        sim_items = [sim_item for sim_item in self.sim_items if sim_item.name in sim_item_names]
        return round(
            sum([sim_item.number_of_examples * (sim_item.receptors_in_repertoire_count if self.is_repertoire else 1)
                 for sim_item in sim_items]))

    def _get_seq_count_for_sim_item(self, signal_id: str, sim_item_name: str) -> int:
        count = 0
        sim_item = [sim_item for sim_item in self.sim_items if sim_item.name == sim_item_name][0]
        signal_with_proportions = {k: v for k, v in sim_item.signal_proportions.items()
                                   if k.id == signal_id or (signal_id in k.id and "__" in k.id)}
        if signal_with_proportions:
            count = sim_item.number_of_examples * sum(signal_with_proportions.values())
            if self.is_repertoire:
                count *= sim_item.receptors_in_repertoire_count
        return count


from typing import List

from sklearn.decomposition import PCA as SklearnPCA

from immuneML.ml_methods.dim_reduction.DimRedMethod import DimRedMethod


class PCA(DimRedMethod):
    """
    Principal component analysis (PCA) method which wraps scikit-learn's PCA. Input arguments for the method are the
    same as supported by scikit-learn (see `PCA scikit-learn documentation
    <https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html#sklearn.decomposition.PCA>`_ for details).

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            ml_methods:
                my_pca:
                    PCA:
                        # arguments as defined by scikit-learn
                        n_components: 2

    """

    def __init__(self, name: str = None, **kwargs):
        super().__init__(name)
        self.method_kwargs = kwargs
        self.method = SklearnPCA(**self.method_kwargs)

    def get_dimension_names(self) -> List[str]:
        return [f"PC{i+1}" for i in range(self.method.n_components)]



import warnings
from pathlib import Path

import pandas as pd

from immuneML.IO.ml_method.UtilIO import UtilIO
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.data_model.EncodedData import EncodedData
from immuneML.data_model.SequenceSet import Repertoire
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.distance_encoding.DistanceMetricType import DistanceMetricType
from immuneML.pairwise_repertoire_comparison.PairwiseRepertoireComparison import PairwiseRepertoireComparison
from immuneML.util import DistanceMetrics
from immuneML.util.EncoderHelper import EncoderHelper
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.ReflectionHandler import ReflectionHandler
from scripts.specification_util import update_docs_per_mapping


class DistanceEncoder(DatasetEncoder):
    """
    Encodes a given RepertoireDataset as distance matrix, where the pairwise distance between each of the repertoires
    is calculated. The distance is calculated based on the presence/absence of elements defined under attributes_to_match.
    Thus, if attributes_to_match contains only 'sequence_aas', this means the distance between two repertoires is maximal
    if they contain the same set of sequence_aas, and the distance is minimal if none of the sequence_aas are shared between
    two repertoires.

    **Specification arguments:**

    - distance_metric (:py:mod:`~immuneML.encodings.distance_encoding.DistanceMetricType`): The metric used to calculate the
      distance between two repertoires. Names of different distance metric types are allowed values in the specification.
      The default distance metric is JACCARD (inverse Jaccard).

    - sequence_batch_size (int): The number of sequences to be processed at once. Increasing this number increases the memory use.
      The default value is 1000.

    - attributes_to_match (list): The attributes to consider when determining whether a sequence is present in both repertoires.
      Only the fields defined under attributes_to_match will be considered, all other fields are ignored.
      Valid values include any repertoire attribute as defined in AIRR rearrangement schema (cdr3_aa, v_call, j_call, etc).

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            encodings:
                my_distance_encoder:
                    Distance:
                        distance_metric: JACCARD
                        sequence_batch_size: 1000
                        attributes_to_match:
                            - cdr3_aa
                            - v_call
                            - j_call

    """

    def __init__(self, distance_metric: DistanceMetricType, attributes_to_match: list, sequence_batch_size: int, context: dict = None,
                 name: str = None):
        super().__init__(name=name)
        self.distance_metric = distance_metric
        self.distance_fn = ReflectionHandler.import_function(self.distance_metric.value, DistanceMetrics)
        self.attributes_to_match = attributes_to_match
        self.sequence_batch_size = sequence_batch_size
        self.context = context
        self.comparison = None

    def set_context(self, context: dict):
        self.context = context
        return self

    @staticmethod
    def _prepare_parameters(distance_metric: str, attributes_to_match: list, sequence_batch_size: int, context: dict = None, name: str = None):
        valid_metrics = [metric.name for metric in DistanceMetricType]
        ParameterValidator.assert_in_valid_list(distance_metric, valid_metrics, "DistanceEncoder", "distance_metric")

        return {
            "distance_metric": DistanceMetricType[distance_metric.upper()],
            "attributes_to_match": attributes_to_match,
            "sequence_batch_size": sequence_batch_size,
            "context": context,
            "name": name
        }

    @staticmethod
    def build_object(dataset, **params):
        if isinstance(dataset, RepertoireDataset):
            prepared_params = DistanceEncoder._prepare_parameters(**params)
            return DistanceEncoder(**prepared_params)
        else:
            raise ValueError("DistanceEncoder is not defined for dataset types which are not RepertoireDataset.")

    def build_distance_matrix(self, dataset: RepertoireDataset, params: EncoderParams, train_repertoire_ids: list):
        self.comparison = PairwiseRepertoireComparison(self.attributes_to_match, self.attributes_to_match, params.result_path,
                                                  sequence_batch_size=self.sequence_batch_size)

        current_dataset = dataset if self.context is None or "dataset" not in self.context else self.context["dataset"]

        distance_matrix = self.comparison.compare(current_dataset, self.distance_fn, self.distance_metric.value)

        repertoire_ids = dataset.get_repertoire_ids()

        distance_matrix = distance_matrix.loc[repertoire_ids, train_repertoire_ids]

        return distance_matrix

    def build_labels(self, dataset: RepertoireDataset, params: EncoderParams) -> dict:

        lbl = ["identifier"]
        lbl.extend(params.label_config.get_labels_by_name())

        tmp_labels = dataset.get_metadata(lbl, return_df=True)
        tmp_labels = tmp_labels.iloc[pd.Index(tmp_labels['identifier']).get_indexer(dataset.get_repertoire_ids())]
        tmp_labels = tmp_labels.to_dict("list")
        del tmp_labels["identifier"]

        return tmp_labels

    def encode(self, dataset, params: EncoderParams) -> RepertoireDataset:
        train_repertoire_ids = EncoderHelper.prepare_training_ids(dataset, params)
        distance_matrix = self.build_distance_matrix(dataset, params, train_repertoire_ids)
        labels = self.build_labels(dataset, params) if params.encode_labels else None

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = EncodedData(examples=distance_matrix, labels=labels, example_ids=distance_matrix.index.values,
                                                   encoding=DistanceEncoder.__name__)

        return encoded_dataset


    @staticmethod
    def load_encoder(encoder_file: Path):
        encoder = DatasetEncoder.load_encoder(encoder_file)
        encoder.comparison = UtilIO.import_comparison_data(encoder_file.parent)
        return encoder

    @staticmethod
    def get_documentation():
        doc = str(DistanceEncoder.__doc__)

        valid_values = [metric.name for metric in DistanceMetricType]
        valid_values = str(valid_values)[1:-1].replace("'", "`")
        mapping = {
            "Names of different distance metric types are allowed values in the specification.": f"Valid values are: {valid_values}.",
        }
        doc = update_docs_per_mapping(doc, mapping)
        return doc



from pathlib import Path
from typing import List

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.hyperparameter_optimization.HPSetting import HPSetting
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.ml_reports.MLReport import MLReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class ConfounderAnalysis(MLReport):
    """
    A report that plots the numbers of false positives and false negatives with respect to each value of
    the metadata features specified by the user. This allows checking whether a given machine learning model makes more
    misclassifications for some values of a metadata feature than for the others.

    **Specification arguments:**

    - metadata_labels (list): A list of the metadata features to use as a basis for the calculations


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_confounder_report:
                    ConfounderAnalysis:
                        metadata_labels:
                          - age
                          - sex

    """

    @classmethod
    def build_object(cls, **kwargs):

        ParameterValidator.assert_keys(kwargs.keys(), ['metadata_labels', 'name'], ConfounderAnalysis.__name__, ConfounderAnalysis.__name__)
        ParameterValidator.assert_type_and_value(kwargs['metadata_labels'], list, ConfounderAnalysis.__name__, 'metadata_labels')
        ParameterValidator.assert_all_type_and_value(kwargs['metadata_labels'], str, ConfounderAnalysis.__name__, 'metadata_labels')
        ParameterValidator.assert_type_and_value(kwargs['name'], str, ConfounderAnalysis.__name__, 'name')

        return ConfounderAnalysis(metadata_labels=kwargs['metadata_labels'], name=kwargs['name'])

    def __init__(self, metadata_labels: List[str], train_dataset: Dataset = None, test_dataset: Dataset = None,
                 method: MLMethod = None, result_path: Path = None, name: str = None, hp_setting: HPSetting = None,
                 label=None, number_of_processes: int=1):
        super().__init__(train_dataset=train_dataset, test_dataset=test_dataset, method=method, result_path=result_path,
                         name=name, hp_setting=hp_setting, label=label, number_of_processes=number_of_processes)
        self.metadata_labels = metadata_labels

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)
        paths = []

        # make predictions
        predictions = self.method.predict(self.test_dataset.encoded_data, self.label)[self.label.name]

        true_labels = self.test_dataset.get_metadata(self.metadata_labels + [self.label.name])
        metrics = ["FP", "FN"]

        plot = make_subplots(rows=len(self.metadata_labels), cols=2)
        listOfPlot = []

        for label_index, meta_label in enumerate(self.metadata_labels):
            csv_data = {}
            for metric_index, metric in enumerate(metrics):
                plotting_data = self._metrics(metric=metric, label_name=self.label.name, meta_label=meta_label,
                                              predictions=predictions, true_labels=true_labels)

                csv_data[f"{metric}"] = plotting_data[f"{metric}"]

                plot.add_trace(go.Bar(x=plotting_data[meta_label], y=plotting_data[metric]), row=label_index + 1, col=metric_index + 1)
                plot.update_xaxes(title_text=f"{meta_label}", row=label_index + 1, col=metric_index + 1, type='category')
                plot.update_yaxes(title_text=f"{metric}", row=label_index + 1, col=metric_index + 1, rangemode="nonnegative", tick0=0, dtick=1)

            csv_data[f"{meta_label}"] = plotting_data[f"{meta_label}"]

            csv_data = pd.DataFrame(csv_data)

            listOfPlot.append(csv_data)

        plot.update_traces(marker_color=px.colors.sequential.Teal[3], showlegend=False)
        filename = self.result_path / "plots.html"
        filename = PlotlyUtil.write_image_to_file(plot, filename)
        report_output_fig = ReportOutput(filename)
        paths.append(report_output_fig)

        result_table_path = self._write_results_table(listOfPlot, self.metadata_labels)
        return ReportResult(name=self.name,
                            info="Plots the numbers of false positives and false negatives with respect to each value of the metadata features specified by the user.",
                            output_figures=paths, output_tables=[ReportOutput(result_table_path[0])])

    def _write_results_table(self, plotting_data, labels):
        filepaths = []
        for label_index, label in enumerate(labels):
            filepath = self.result_path / f"{label}.csv"
            plotting_data[label_index].to_csv(filepath, index=False)
            filepaths.append(filepath)
        return filepaths

    @staticmethod
    def _metrics(metric, label_name, meta_label, predictions, true_labels):
        # indices of samples at which misclassification occurred
        if metric == "FP":
            metric_inds = np.nonzero(np.greater(predictions, true_labels[label_name]))[0].tolist()
        else:
            metric_inds = np.nonzero(np.less(predictions, true_labels[label_name]))[0].tolist()

        metadata_values = true_labels[meta_label]
        # indices of misclassification with respect to the metadata label
        label_inds = np.array(metadata_values)[metric_inds]

        metric_vals = []
        unique_levels = np.unique(metadata_values)

        # number of metric occurrences at each metadata level
        for val in unique_levels:
            metric_vals.append(np.count_nonzero(label_inds == val))

        plotting_data = pd.DataFrame(
            {f"{metric}": metric_vals, f"{meta_label}": unique_levels})

        return plotting_data



import pickle
import shutil
from pathlib import Path

from immuneML.pairwise_repertoire_comparison.ComparisonData import ComparisonData
from immuneML.pairwise_repertoire_comparison.PairwiseRepertoireComparison import PairwiseRepertoireComparison


class UtilIO:

    @staticmethod
    def export_pairwise_comparison(pairwise_comparison: PairwiseRepertoireComparison, path: Path) -> Path:
        pairwise_comparison_file = path / "pairwise_repertoire_comparison.pickle"
        with pairwise_comparison_file.open("wb") as file:
            pickle.dump(pairwise_comparison, file)

        UtilIO.export_comparison_data(pairwise_comparison.comparison_data, path)

        return pairwise_comparison_file

    @staticmethod
    def import_pairwise_comparison(path: Path) -> PairwiseRepertoireComparison:
        filename = path / "pairwise_repertoire_comparison.pickle"
        with filename.open("rb") as file:
            pairwise_comparison = pickle.load(file)

        pairwise_comparison.comparison_data = UtilIO.import_comparison_data(path)
        pairwise_comparison.path = path

        return pairwise_comparison

    @staticmethod
    def import_comparison_data(path: Path) -> ComparisonData:
        comp_data_path = path / "comp_data"

        comp_data_file = path / "comparison_data.pickle"
        with comp_data_file.open("rb") as file:
            comp_data = pickle.load(file)

        comp_data.path = comp_data_path
        for batch in comp_data.batches:
            batch.path = comp_data_path

        return comp_data

    @staticmethod
    def export_comparison_data(comp_data: ComparisonData, path: Path):
        shutil.copytree(comp_data.path, path / "comp_data")

        comp_data_file = path / "comparison_data.pickle"
        with comp_data_file.open('wb') as file:
            pickle.dump(comp_data, file)

        return comp_data_file



import logging
import os
import random

import numpy as np
import pandas as pd

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.ElementDataset import ElementDataset
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.util.ReflectionHandler import ReflectionHandler
from immuneML.workflows.steps.data_splitter.DataSplitterParams import DataSplitterParams
from immuneML.workflows.steps.data_splitter.Util import Util


class ManualSplitter:

    @staticmethod
    def split_dataset(input_params: DataSplitterParams):
        if isinstance(input_params.dataset, RepertoireDataset):
            return ManualSplitter._split_dataset(input_params, ManualSplitter._make_repertoire_dataset)
        elif isinstance(input_params.dataset, ElementDataset):
            return ManualSplitter._split_dataset(input_params, ManualSplitter._make_element_dataset)
        else:
            raise ValueError(f"DataSplitter: dataset is unexpected class: {type(input_params.dataset).__name__}, "
                             f"expected one of {str(ReflectionHandler.all_nonabstract_subclass_basic_names(Dataset, '', 'dataset/'))[1:-1]}")

    @staticmethod
    def _split_dataset(input_params, make_dataset_func):
        train_metadata_path = input_params.split_config.manual_config.train_metadata_path
        test_metadata_path = input_params.split_config.manual_config.test_metadata_path

        train_dataset = make_dataset_func(input_params, train_metadata_path, Dataset.TRAIN)
        test_dataset = make_dataset_func(input_params, test_metadata_path, Dataset.TEST)

        return [train_dataset], [test_dataset]

    @staticmethod
    def _make_element_dataset(input_params, metadata_path, dataset_type: str) -> ElementDataset:
        example_ids = input_params.dataset.get_example_ids()
        return ManualSplitter._make_subset(input_params, metadata_path, dataset_type, example_ids, 'example_id')

    @staticmethod
    def _make_repertoire_dataset(input_params, metadata_path, dataset_type: str) -> RepertoireDataset:
        subject_ids = input_params.dataset.get_metadata(["subject_id"])["subject_id"]
        return ManualSplitter._make_subset(input_params, metadata_path, dataset_type, subject_ids, 'subject_id')

    @staticmethod
    def _make_subset(input_params, metadata_path, dataset_type, example_ids, col_name):
        ManualSplitter._check_unique_count(example_ids, input_params.dataset)

        metadata_df = ManualSplitter._get_metadata(metadata_path, dataset_type, col_name)
        indices_of_interest = metadata_df[col_name].astype(str).values.tolist()
        indices = [i for i in range(len(example_ids)) if str(example_ids[i]) in indices_of_interest]
        random.shuffle(indices)

        logging.info(f"{ManualSplitter.__name__}: Making {dataset_type} dataset subset with {len(indices)} elements.")

        return Util.make_dataset(input_params.dataset, indices, input_params, 0, dataset_type)

    @staticmethod
    def _check_unique_count(example_ids: list, dataset):
        unique_example_count = np.unique(example_ids).shape[0]
        assert len(example_ids) == unique_example_count, f"DataSplitter: there are {len(example_ids)} elements, but {unique_example_count} " \
                                                         f"unique identifiers. Check the metadata for the original dataset {dataset.name}."

    @staticmethod
    def _get_metadata(metadata_path, dataset_type: str, col_name: str) -> pd.DataFrame:
        metadata_df = pd.read_csv(metadata_path)
        assert col_name in metadata_df, f"DataSplitter: {dataset_type} metadata {os.path.basename(metadata_path)} is missing column " \
                                        f"'{col_name}' which should be used for matching examples when splitting to train and test data."
        return metadata_df



import logging
from pathlib import Path

import pandas as pd
import plotly
import plotly.express as px

from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.dsl.definition_parsers.MLParser import MLParser
from immuneML.ml_methods.dim_reduction.DimRedMethod import DimRedMethod
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.clustering_method_reports.ClusteringMethodReport import ClusteringMethodReport
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.clustering.clustering_run_model import ClusteringItem


class ClusteringVisualization(ClusteringMethodReport):
    """
    A report that creates low-dimensional visualizations of clustering results using the specified dimensionality reduction method.
    For each dataset and clustering configuration, it creates a scatter plot where points are colored by their cluster assignments.

    Specification arguments:

        - dim_red_method (dict): specification of which dimensionality reduction to perform; valid options are presented
          under :ref:`**Dimensionality reduction methods**` and should be specified with the name of the method and its
          parameters, see the example below; if not specified, the report will use any dimensionality reduced data
          present in the dataset's encoded data; if the dataset does not contain dimensionality reduced data, and the
          encoded data has more than 2 dimensions, the report will be skipped.

    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        reports:
            my_report_with_pca:
                ClusteringVisualization:
                    dim_red_method:
                        PCA:
                            n_components: 2
            my_report_with_tsne:
                ClusteringVisualization:
                    dim_red_method:
                        TSNE:
                            n_components: 2
                            init: pca
            my_report_existing_dim_red:
                ClusteringVisualization:
                    dim_red_method: null

    """

    def __init__(self, dim_red_method: DimRedMethod = None, name: str = None,
                 result_path: Path = None, clustering_item: ClusteringItem = None):
        super().__init__(name=name, result_path=result_path, clustering_item=clustering_item)
        self.dim_red_method = dim_red_method
        self.result_name = None
        self.desc = "Clustering Visualization"
        self._dimension_names = self.dim_red_method.get_dimension_names() if self.dim_red_method else None

    @classmethod
    def build_object(cls, **kwargs):
        location = "ClusteringVisualization"
        name = kwargs["name"] if "name" in kwargs else None
        result_path = kwargs["result_path"] if "result_path" in kwargs else None

        if "dim_red_method" in kwargs and kwargs["dim_red_method"]:
            method_name = list(kwargs["dim_red_method"].keys())[0]
            dim_red_method = MLParser.parse_any_model("dim_red_method", kwargs["dim_red_method"], method_name)[0]
        else:
            logging.warning(f"{location}: No dimensionality reduction method specified. "
                            "If the encoded dataset includes dimensionality reduction, it will be used.")
            dim_red_method = None

        return cls(dim_red_method=dim_red_method, name=name, result_path=result_path,
                   clustering_item=kwargs['clustering_item'] if 'clustering_item' in kwargs else None,)

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)
        self.result_name = f"clustering_{self.dim_red_method.__class__.__name__.lower()}_plots"
        result_path = PathBuilder.build(self.result_path / self.result_name)

        plot_path = self._make_plot(result_path)
        report_output = ReportOutput(plot_path,
                                     f"Clustering visualization for {self.item.cl_setting.get_key()}")

        return ReportResult(f"{self.desc} ({self.name})",
                            info=f"Visualizations of clustering results",
                            output_figures=[report_output])

    def _make_plot(self, result_path: Path) -> Path:
        if self.dim_red_method is not None:
            transformed_data = self.dim_red_method.fit_transform(dataset=self.item.dataset)
        elif self.item.dataset.encoded_data.dimensionality_reduced_data is not None:
            transformed_data = self.item.dataset.encoded_data.dimensionality_reduced_data
            self._dimension_names = self.item.dataset.encoded_data.dim_names if self.item.dataset.encoded_data.dim_names else ['dim1', 'dim2']
        elif self.item.dataset.encoded_data.examples.shape[1] <= 2:
            transformed_data = self.item.dataset.encoded_data.get_examples_as_np_matrix()
            self._dimension_names = self.item.dataset.encoded_data.feature_names
        else:
            raise ValueError("ClusteringVisualization: No dimensionality reduction method specified, and the dataset "
                             "does not contain dimensionality reduced data. Please specify a dimensionality reduction "
                             "method.")

        df = pd.DataFrame(transformed_data, columns=self._dimension_names)
        df['cluster'] = pd.Series(self.item.predictions).astype(str)
        df['id'] = self.item.dataset.get_example_ids()

        fig = px.scatter(df, x=self._dimension_names[0], y=self._dimension_names[1], color='cluster',
                         color_discrete_sequence=plotly.colors.qualitative.Set2,
                         category_orders={'cluster': sorted(df.cluster.unique())},
                         hover_data=['id'])

        fig.update_layout(template="plotly_white")

        df.to_csv(result_path / f"clustering_visualization_{self.dim_red_method.name if self.dim_red_method else ''}.csv", index=False)

        plot_path = PlotlyUtil.write_image_to_file(fig,
                                                   result_path / f"clustering_visualization_{self.dim_red_method.name if self.dim_red_method else ''}.html",
                                                   df.shape[0])

        return plot_path

    def get_ids(self):
        if isinstance(self.item.dataset, RepertoireDataset):
            metadata = self.item.dataset.get_metadata(['subject_id'], return_df=True)
            if 'subject_id' in metadata.columns:
                return metadata['subject_id'].tolist()
            else:
                return self.item.dataset.get_example_ids()
        else:
            return self.item.dataset.get_example_ids()



import sys
import os
import yaml
from pathlib import Path
import logging
import warnings
import inspect
from immuneML.util.PathBuilder import PathBuilder

def setup_logger(log_file):
    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(levelname)s: %(message)s',
                        handlers=[logging.FileHandler(log_file, mode="w"),
                                  logging.StreamHandler(stream=sys.stdout)])
    warnings.showwarning = lambda message, category, filename, lineno, file=None, line=None: logging.warning(message)

def set_tmp_path(tmp_path):
    PathBuilder.remove_old_and_build(tmp_path)
    logging.info(f"Storing temporary files at '{tmp_path}'")

def check_is_alphanum_name(filename_stem):
    for char in filename_stem:
        assert char.isalnum(), f"Error: class name is only allowed to contain alphanumeric characters, found: {char}. Please rename the encoder (both class and filename) to an alphanumeric name."

def load_default_params_file(default_params_filepath):
    logging.info("Checking default parameters file...")

    assert os.path.isfile(default_params_filepath), f"Error: default parameters file was expected to be found at: '{default_params_filepath}'\n" \
                                                    f"To test new classes without default parameters, please run this script with the '-p' flag. "
    logging.info(f"...Default parameters file was found at: '{default_params_filepath}'")

    logging.info("Attempting to load default parameters file with yaml.load()...")
    with Path(default_params_filepath).open("r") as file:
        default_params = yaml.load(file, Loader=yaml.FullLoader)

    logging.info(f"...The following default parameters were found: '{default_params}'")

    return default_params

def check_default_args(default_params_filepath, no_default_parameters, class_name):
    if no_default_parameters:
        if os.path.isfile(default_params_filepath):
            logging.warning(f"Default parameters file was found at '{default_params_filepath} but 'no_default_parameters' flag was enabled. "
                            f"Assuming no default parameters for further testing, but note that default parameters will be attempted to be loaded by immuneML when using {class_name}.")
        else:
            logging.info("Skip default parameter testing as 'no_default_parameters' flag is enabled")
        return {}
    else:
        return load_default_params_file(default_params_filepath)


def check_base_vs_instance_methods(base_class, subclass_instance):
    base_class_functions = [name for name, _ in inspect.getmembers(base_class, predicate=inspect.isfunction) if
                            not name.startswith("_")]
    class_methods = [name for name, _ in inspect.getmembers(subclass_instance, predicate=inspect.ismethod) if
                     not name.startswith("_")]
    additional_public_methods = [name for name in class_methods if name not in base_class_functions]

    if len(additional_public_methods) > 0:
        logging.warning(
            f"In immuneML, 'public' methods start without underscore while 'private' methods start with underscore. "
            f"Methods added to specific encoders other than those inherited from {base_class.__name__} should generally be considered private, although this is not strictly enforced. "
            f"The following additional public methods were found: {', '.join(additional_public_methods)}. To remove this warning, rename these methods to start with _underscore.")

    for method_name in class_methods:
        if method_name != method_name.lower():
            logging.warning(f"Method names must be written in snake_case, found the following method: {method_name}. Please rename this method.")




import re


class FilenameHandler:

    @staticmethod
    def _to_snake_case(string: str):
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', re.sub('(.)([A-Z][a-z]+)', r'\1_\2', string)).lower()

    @staticmethod
    def get_filename(class_name: str, file_type: str):
        """
        converts the class name to snake case and appends given file type
        :param class_name: name of the class that will be stored in the file
        :param file_type: file extension: pickle, json
        :return: filename consisting of concatenated class_name in snake case and file type
        """
        name = FilenameHandler._to_snake_case(class_name)
        if file_type != "":
            name += ".{}".format(file_type)

        return name

    @staticmethod
    def get_dataset_name(class_name: str):
        if "Encoder" in class_name:
            name = "encoded_dataset.yaml"
        else:
            name = "dataset.yaml"

        return name

    @staticmethod
    def get_model_name(class_name: str, file_type: str = "pickle"):
        name = FilenameHandler._to_snake_case(class_name) + "_model." + file_type
        return name



import re
from itertools import zip_longest
import numpy as np
import logging
from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.ElementDataset import ReceptorDataset
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.protein_embedding.ProteinEmbeddingEncoder import ProteinEmbeddingEncoder
from immuneML.util.NumpyHelper import NumpyHelper
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.Logger import log_memory_usage


class TCRBertEncoder(ProteinEmbeddingEncoder):
    """
    TCRBertEncoder is based on `TCR-BERT <https://github.com/wukevin/tcr-bert/tree/main>`_, a large language model
    trained on TCR sequences. TCRBertEncoder embeds TCR sequences using either of the pre-trained models provided on
    HuggingFace repository.

    Original publication:
    Wu, K. E., Yost, K., Daniel, B., Belk, J., Xia, Y., Egawa, T., Satpathy, A., Chang, H., & Zou, J. (2024).
    TCR-BERT: Learning the grammar of T-cell receptors for flexible antigen-binding analyses. Proceedings of the
    18th Machine Learning in Computational Biology Meeting, 194–229. https://proceedings.mlr.press/v240/wu24b.html

    **Dataset type:**

    - SequenceDataset

    - ReceptorDataset

    - RepertoireDataset

    **Specification arguments:**

    - model (str): The pre-trained model to use (huggingface model hub identifier). Available options are 'tcr-bert'
      and 'tcr-bert-mlm-only'.

    - layers (list): The hidden layers to use for encoding. Layers should be given as negative integers, where -1
      indicates the last representation, -2 second to last, etc. Default is [-1].

    - method (str): The method to use for pooling the hidden states. Available options are 'mean', 'max'',
      'cls', and 'pool'. Default is 'mean'. For explanation of the methods, see GitHub repository of TCR-BERT.

    - batch_size (int): The number of sequences to encode at the same time. This could have large impact on memory usage.
      If memory is an issue, try with smaller batch sizes. Defaults to 4096.

    - scale_to_zero_mean (bool): Whether to scale the embeddings to zero mean. Defaults to True.

    - scale_to_unit_variance (bool): Whether to scale the embeddings to unit variance. Defaults to True.

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            encodings:
                my_tcr_bert_encoder: TCRBert

    """
    def __init__(self, name: str = None, region_type: RegionType = RegionType.IMGT_CDR3, model: str = None,
                 layers: list = None, method: str = None, batch_size: int = None, device: str = 'cpu',
                 scale_to_zero_mean: bool = True, scale_to_unit_variance: bool = True):
        super().__init__(region_type, name, num_processes=1, device=device, batch_size=batch_size,
                         scale_to_zero_mean=scale_to_zero_mean, scale_to_unit_variance=scale_to_unit_variance)
        self.model = model
        self.layers = layers
        self.method = method
        self.embedding_dim = 768

    @staticmethod
    def build_object(dataset: Dataset, **params):
        prepared_params = TCRBertEncoder._prepare_parameters(**params)
        return TCRBertEncoder(**prepared_params)

    @staticmethod
    def _prepare_parameters(name: str = None, model: str = None, layers: list = None, method: str = None,
                            batch_size: int = None, device: str = None):
        location = TCRBertEncoder.__name__
        ParameterValidator.assert_in_valid_list(model, ["tcr-bert", "tcr-bert-mlm-only"], location, "model")
        ParameterValidator.assert_type_and_value(layers, list, location, "layers")
        ParameterValidator.assert_in_valid_list(method, ["mean", "max", "attn_mean", "cls", "pool"], location, "method")
        ParameterValidator.assert_type_and_value(batch_size, int, location, "batch_size")
        ParameterValidator.assert_type_and_value(device, str, location, 'device')
        if len(re.findall("cuda:[0-9]*", device)) == 0:
            ParameterValidator.assert_in_valid_list(device, ['cpu', 'mps', 'cuda'], location, 'device')
        return {"name": name, "model": model, "layers": layers, "method": method, "batch_size": batch_size,
                'device': device}

    def _get_caching_params(self, dataset, params: EncoderParams, step: str = None):
        return (dataset.identifier, tuple(params.label_config.get_labels_by_name()), self.scale_to_zero_mean,
                self.scale_to_unit_variance, step, self.region_type.name, tuple(self.layers), self._get_encoding_name(),
                params.learn_model)

    def _get_model_and_tokenizer(self, log_location):
        from transformers import BertModel, BertTokenizer
        
        log_memory_usage(stage="start", location=log_location)
        logging.info(f"TCRBert ({self.name}): Loading model: wukevin/{self.model}")
        
        model = BertModel.from_pretrained(f"wukevin/{self.model}", attn_implementation="eager")
        log_memory_usage("after model load", log_location)
        
        model = model.to(self.device).eval()
        log_memory_usage("after model to device", log_location)
        
        tokenizer = BertTokenizer.from_pretrained(
            f"wukevin/{self.model}",
            do_basic_tokenize=False,
            do_lower_case=False,
            tokenize_chinese_chars=False,
            unk_token="?",
            sep_token="|",
            pad_token="$",
            cls_token="*",
            mask_token=".",
            padding_side="right"
        )
        log_memory_usage("after tokenizer load", log_location)
        
        return model, tokenizer

    def _embed_sequence_set(self, sequence_set, seq_field):
        import torch
        
        log_location = f"TCRBertEncoder ({self.name})"
        model, tokenizer = self._get_model_and_tokenizer(log_location)
        
        seqs = self._get_sequences(sequence_set, seq_field)
        n_sequences = len(seqs)
        
        # Calculate embedding dimension based on number of layers
        total_dim = len(self.layers) * self.embedding_dim if self.method != "pool" else self.embedding_dim
        
        # Create memory-mapped array for embeddings
        embeddings = NumpyHelper.create_memmap_array_in_cache((n_sequences, total_dim))
        
        chunks = [seqs[i: i + self.batch_size] for i in range(0, n_sequences, self.batch_size)]
        chunks_pair = [None] * len(chunks)  # Create matching None pairs for zip_longest
        chunks_zipped = list(zip_longest(chunks, chunks_pair))
        
        current_idx = 0
        with torch.no_grad():
            for batch_idx, seq_chunk in enumerate(chunks_zipped):
                logging.info(
                    f"{log_location}: Processing batch {batch_idx + 1}/{len(chunks_zipped)}"
                )
                
                encoded = tokenizer(
                    *seq_chunk, padding="max_length", max_length=64, return_tensors="pt"
                )
                encoded = {k: v.to(self.device) for k, v in encoded.items()}
                
                x = model.forward(**encoded, output_hidden_states=True, output_attentions=True)
                
                if self.method == "pool":
                    batch_size = len(seq_chunk[0])
                    embeddings[current_idx:current_idx + batch_size] = x.pooler_output.cpu().numpy()
                    current_idx += batch_size
                else:
                    batch_embeddings = []
                    for i in range(len(seq_chunk[0])):
                        e = []
                        for l in self.layers:
                            h = x.hidden_states[l][i].cpu().numpy()
                            
                            if self.method == "cls":
                                e.append(h[0])
                                continue
                                
                            if seq_chunk[1] is None:
                                seq_len = len(seq_chunk[0][i].split())
                            else:
                                seq_len = len(seq_chunk[0][i].split()) + len(seq_chunk[1][i].split()) + 1
                                
                            seq_hidden = h[1: 1 + seq_len]
                            
                            if self.method == "mean":
                                e.append(seq_hidden.mean(axis=0))
                            elif self.method == "max":
                                e.append(seq_hidden.max(axis=0))
                            else:
                                raise ValueError(f"Unrecognized method: {self.method}")
                                
                        e = np.hstack(e)
                        batch_embeddings.append(e)
                    
                    batch_embeddings = np.stack(batch_embeddings)
                    embeddings[current_idx:current_idx + len(batch_embeddings)] = batch_embeddings
                    current_idx += len(batch_embeddings)
                
                del x, encoded
                if torch.cuda.is_available():
                    torch.cuda.empty_cache()
                
                log_memory_usage(f"after batch {batch_idx + 1}", log_location)
        
        logging.info(f"{log_location}: Finished processing all sequences")
        return embeddings

    def _get_sequences(self, sequence_set, field_name):
        seqs = getattr(sequence_set, field_name).tolist()
        seqs = [" ".join(list(s)) for s in seqs]
        return seqs

    def _get_encoding_name(self) -> str:
        return f"TCRBertEncoder({self.model})"

    def _get_model_link(self) -> str:
        return self.model



from pathlib import Path
from typing import Union

from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.presentation.TemplateParser import TemplateParser
from immuneML.presentation.html.Util import Util
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.ligo_sim_feasibility.FeasibilitySummaryInstruction import FeasibilitySummaryState


class FeasibilitySummaryHTMLBuilder:
    CSS_PATH = EnvironmentSettings.html_templates_path / "css/custom.css"

    @staticmethod
    def build(state: FeasibilitySummaryState) -> Path:
        """
        Function that builds the HTML files based on the ExploratoryAnalysis state.
        Arguments:
            state: ExploratoryAnalysisState object with details and results of the instruction
        Returns:
             path to the main HTML file (which is located under state.result_path)
        """
        base_path = PathBuilder.build(state.result_path / "../HTML_output/")
        html_map = FeasibilitySummaryHTMLBuilder.make_html_map(state, base_path)
        result_file = base_path / f"FeasibilitySummary_{state.name}.html"

        TemplateParser.parse(template_path=EnvironmentSettings.html_templates_path / "FeasibilitySummary.html",
                             template_map=html_map, result_path=result_file)

        return result_file

    @staticmethod
    def make_html_map(state: FeasibilitySummaryState, base_path: Path) -> dict:
        html_map = {
            "css_style": Util.get_css_content(FeasibilitySummaryHTMLBuilder.CSS_PATH),
            "full_specs": Util.get_full_specs_path(base_path),
            "logfile": Util.get_logfile_path(base_path),
            "sequence_count": state.sequence_count,
            "signal_names": ", ".join(s.id for s in state.signals),
            "gen_models": [{
                "name": name,
                "sig_freq": make_from_report_result(reports.signal_frequencies, base_path),
                "sig_coocc": make_from_report_result(reports.signal_cooccurrences, base_path),
                "p_gens": make_from_report_result(reports.p_gen_histogram, base_path),
                "seq_len": make_from_report_result(reports.seq_len_dist, base_path)
            } for name, reports in state.reports.items()]
        }

        return html_map


def make_from_report_result(report_result, base_path) -> Union[dict, None]:
    if report_result is not None:
        report_dict = Util.to_dict_recursive(Util.update_report_paths(report_result, base_path), base_path)
        report_dict['show_info'] = report_dict['info'] != '' and report_dict['info'] != None
        report_dict['show_text'] = len(report_dict['output_text']) > 0
        report_dict['show_tables'] = len(report_dict['output_tables']) > 0
        report_dict['show_figures'] = len(report_dict['output_figures']) > 0
        return report_dict
    else:
        return None



from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.steps.data_splitter.DataSplitterParams import DataSplitterParams


class Util:

    @staticmethod
    def prepare_path(input_params: DataSplitterParams, split_index: int, dataset_type: str) -> str:
        path = input_params.paths[split_index] / f"datasets/{dataset_type.lower()}/"
        PathBuilder.build(path)
        return path

    @staticmethod
    def make_dataset(dataset: Dataset, indices, input_params: DataSplitterParams, i: int, dataset_type: str):
        path = Util.prepare_path(input_params, i, dataset_type)
        new_dataset = dataset.make_subset(indices, path, dataset_type)
        return new_dataset



from scipy.stats import lognorm
import pandas as pd

import logging
import plotly.express as px
import plotly.graph_objects as go

from immuneML.encodings.motif_encoding.PositionalMotifHelper import PositionalMotifHelper
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput


class MotifPerformancePlotHelper():

    @staticmethod
    def get_plotting_data(training_encoded_data, test_encoded_data, highlight_motifs_path=None, highlight_motifs_name="highlight"):
        training_feature_annotations = MotifPerformancePlotHelper._get_annotated_feature_annotations(training_encoded_data, highlight_motifs_path, highlight_motifs_name)
        test_feature_annotations = MotifPerformancePlotHelper._get_annotated_feature_annotations(test_encoded_data, highlight_motifs_path, highlight_motifs_name)

        training_feature_annotations["training_TP"] = training_feature_annotations["TP"]
        test_feature_annotations = MotifPerformancePlotHelper.merge_train_test_feature_annotations(training_feature_annotations, test_feature_annotations)

        return training_feature_annotations, test_feature_annotations

    @staticmethod
    def _get_annotated_feature_annotations(encoded_data, highlight_motifs_path, highlight_motifs_name):
        feature_annotations = encoded_data.feature_annotations.copy()
        MotifPerformancePlotHelper._annotate_confusion_matrix(feature_annotations)
        MotifPerformancePlotHelper._annotate_highlight(feature_annotations, highlight_motifs_path, highlight_motifs_name)

        return feature_annotations

    @staticmethod
    def _annotate_confusion_matrix(feature_annotations):
        feature_annotations["precision"] = feature_annotations.apply(
            lambda row: 0 if row["TP"] == 0 else row["TP"] / (row["TP"] + row["FP"]), axis="columns")

        feature_annotations["recall"] = feature_annotations.apply(
            lambda row: 0 if row["TP"] == 0 else row["TP"] / (row["TP"] + row["FN"]), axis="columns")

    @staticmethod
    def _annotate_highlight(feature_annotations, highlight_motifs_path, highlight_motifs_name):
        feature_annotations["highlight"] = MotifPerformancePlotHelper._get_highlight(feature_annotations, highlight_motifs_path, highlight_motifs_name)

    @staticmethod
    def _get_highlight(feature_annotations, highlight_motifs_path, highlight_motifs_name):
        if highlight_motifs_path is not None:
            # highlight_motifs = [PositionalMotifHelper.motif_to_string(indices, amino_acids, motif_sep="-", newline=False)
            #                     for indices, amino_acids in PositionalMotifHelper.read_motifs_from_file(highlight_motifs_path)]

            highlight_motifs = PositionalMotifHelper.read_motifs_from_file(highlight_motifs_path)
            motifs = [PositionalMotifHelper.string_to_motif(motif, value_sep="&", motif_sep="-") for motif in feature_annotations["feature_names"]]

            return [highlight_motifs_name if MotifPerformancePlotHelper._is_highlight_motif(motif, highlight_motifs) else "Motif"
                    for motif in motifs]
        else:
            return ["Motif"] * len(feature_annotations)

    @staticmethod
    def _is_highlight_motif(motif, highlight_motifs):
        for highlight_motif in highlight_motifs:
            if motif == highlight_motif:
                return True

            if len(motif[0]) > len(highlight_motif[0]):
                if MotifPerformancePlotHelper.is_sub_motif(highlight_motif, motif):
                    return True

        return False

    @staticmethod
    def is_sub_motif(short_motif, long_motif):
        assert len(long_motif[0]) > len(short_motif[0])

        long_motif_dict = {long_motif[0][i]: long_motif[1][i] for i in range(len(long_motif[0]))}

        for idx, aa in zip(short_motif[0], short_motif[1]):
            if idx in long_motif_dict.keys():
                if long_motif_dict[idx] != aa:
                    return False
            else:
                return False

        return True


    @staticmethod
    def merge_train_test_feature_annotations(training_feature_annotations, test_feature_annotations):
        training_info_to_merge = training_feature_annotations[["feature_names", "training_TP"]].copy()
        test_info_to_merge = test_feature_annotations.copy()

        merged_train_test_info = training_info_to_merge.merge(test_info_to_merge)

        return merged_train_test_info

    @staticmethod
    def get_combined_precision(plotting_data, min_points_in_window, smoothing_constant1, smoothing_constant2):
        group_by_tp = plotting_data.groupby("training_TP")

        combined_precision = group_by_tp["TP"].sum() / (group_by_tp["TP"].sum() + group_by_tp["FP"].sum())

        df = pd.DataFrame({"training_TP": list(combined_precision.index),
                           "combined_precision": list(combined_precision)})

        df["smooth_combined_precision"] = MotifPerformancePlotHelper._smooth_combined_precision(list(combined_precision.index),
                                                                                                    list(combined_precision),
                                                                                                    list(group_by_tp["TP"].count()),
                                                                                                    min_points_in_window,
                                                                                                    smoothing_constant1,
                                                                                                    smoothing_constant2)

        return df

    @staticmethod
    def _smooth_combined_precision(x, y, weights, min_points_in_window, smoothing_constant1, smoothing_constant2):
        smoothed_y = []

        for i in range(len(x)):
            scale = MotifPerformancePlotHelper._get_lognorm_scale(x, i, weights, min_points_in_window, smoothing_constant1, smoothing_constant2)

            lognorm_for_this_x = lognorm.pdf(x, s=0.1, loc=x[i] - scale, scale=scale)

            smoothed_y.append(sum(lognorm_for_this_x * y) / sum(lognorm_for_this_x))

        return smoothed_y

    @staticmethod
    def _get_lognorm_scale(x, i, weights, min_points_in_window, smoothing_constant1, smoothing_constant2):
        window_size = MotifPerformancePlotHelper._determine_window_size(x, i, weights, min_points_in_window)
        return window_size * smoothing_constant1 + smoothing_constant2

    @staticmethod
    def _determine_window_size(x, i, weights, min_points_in_window):
        x_rng = 0
        n_data_points = weights[i]

        if sum(weights) < min_points_in_window:
            logging.warning(f"{MotifPerformancePlotHelper.__name__}: min_points_in_window ({min_points_in_window}) is smaller than the total number of points in the plot ({sum(weights)}). Setting min_points_in_window to {sum(weights)} instead...")
            min_points_in_window = sum(weights)
        else:
            min_points_in_window = min_points_in_window

        while n_data_points < min_points_in_window:
            x_rng += 1

            to_select = [j for j in range(len(x)) if (x[i] - x_rng) <= x[j] <= (x[i] + x_rng)]
            lower_index = min(to_select)
            upper_index = max(to_select)

            n_data_points = sum(weights[lower_index:upper_index + 1])

        return x_rng

    @staticmethod
    def plot_precision_per_tp(file_path, plotting_data, combined_precision, dataset_type, training_set_name,
                              tp_cutoff, motifs_name="motifs", highlight_motifs_name="highlight"):
        # fig = px.scatter(plotting_data,
        #                y="precision", x="training_TP", hover_data=["feature_names"],
        #                range_y=[0, 1.01], color_discrete_sequence=["#74C4C4"],
        #                # stripmode="overlay",
        #                log_x=True,
        #                labels={
        #                    "precision": f"Precision ({dataset_type})",
        #                    "feature_names": "Motif",
        #                    "training_TP": f"True positive predictions ({training_set_name})"
        #                }, template="plotly_white")


        # make 'base figure' with 1 point
        fig = px.scatter(plotting_data, y=[0], x=[0], range_y=[-0.01, 1.01], log_x=True,
                         template="plotly_white")

        # hide 'base figure' point
        fig.update_traces(marker=dict(size=12, opacity=0), selector=dict(mode='markers'))

        # add data points (needs to be separate trace to show up in legend)
        fig.add_trace(go.Scatter(x=plotting_data["training_TP"], y=plotting_data["precision"],
                                 mode='markers', name="Motif precision",
                                 marker=dict(symbol="circle", color="#74C4C4")),
                      secondary_y=False)

        # add combined precision
        fig.add_trace(go.Scatter(x=combined_precision["training_TP"], y=combined_precision["combined_precision"],
                                 mode='markers+lines', name="Combined precision",
                                 marker=dict(symbol="diamond", color=px.colors.diverging.Tealrose[0])),
                      secondary_y=False)

        # add highlighted motifs
        plotting_data_highlight = plotting_data[plotting_data["highlight"] != "Motif"]
        if len(plotting_data_highlight) > 0:
            fig.add_trace(go.Scatter(x=plotting_data_highlight["training_TP"], y=plotting_data_highlight["precision"],
                                     mode='markers', name=f"{highlight_motifs_name} precision",
                                     marker=dict(symbol="circle", color="#F5C144")),
                          secondary_y=False)

        # add smoothed combined precision
        if "smooth_combined_precision" in combined_precision:
            fig.add_trace(go.Scatter(x=combined_precision["training_TP"], y=combined_precision["smooth_combined_precision"],
                                     marker=dict(color=px.colors.diverging.Tealrose[-1]),
                                     name="Combined precision, smoothed",
                                     mode="lines", line_shape='spline', line={'smoothing': 1.3}),
                          secondary_y=False, )

        # add vertical TP cutoff line
        if tp_cutoff is not None:
            if tp_cutoff == "auto":
                tp_cutoff = min(plotting_data["training_TP"])

            fig.add_vline(x=tp_cutoff, line_dash="dash")

        tickvals = MotifPerformancePlotHelper._get_log_x_axis_ticks(plotting_data, tp_cutoff)
        fig.update_layout(xaxis=dict(tickvals=tickvals),
                          xaxis_title=f"True positive predictions ({training_set_name})",
                          yaxis_title=f"Precision ({dataset_type})",
                          showlegend=True)

        file_path = PlotlyUtil.write_image_to_file(fig, file_path)

        return ReportOutput(
            path=file_path,
            name=f"Precision scores on the {dataset_type} for {motifs_name} found at each true positive count of the {training_set_name}.",
        )

    @staticmethod
    def _get_log_x_axis_ticks(plotting_data, tp_cutoff):
        ticks = []

        min_val, max_val = min(plotting_data["training_TP"]), max(plotting_data["training_TP"])

        i = 1
        while i < max_val:
            if i > min_val:
                ticks.append(i)
            i *= 10

        ticks.append(min_val)
        ticks.append(max_val)

        if tp_cutoff is not None:
            ticks.append(tp_cutoff)

        return sorted(ticks)

    @staticmethod
    def plot_precision_recall(file_path, plotting_data, min_recall=None, min_precision=None, dataset_type=None, motifs_name="motifs",
                              highlight_motifs_name="highlight"):
        fig = px.scatter(plotting_data,
                         y="precision", x="recall", hover_data=["feature_names"],
                         range_x=[0, 1.01], range_y=[0, 1.01], color="highlight",
                         color_discrete_map={"Motif": px.colors.qualitative.Pastel[0],
                                             highlight_motifs_name: px.colors.qualitative.Pastel[1]},
                         labels={
                             "precision": f"Precision ({dataset_type})",
                             "recall": f"Recall ({dataset_type})",
                             "feature_names": "Motif",
                         }, template="plotly_white")

        if min_precision is not None and min_precision > 0:
            fig.add_hline(y=min_precision, line_dash="dash")

        if min_recall is not None and min_recall > 0:
            fig.add_vline(x=min_recall, line_dash="dash")

        file_path = PlotlyUtil.write_image_to_file(fig, file_path)

        return ReportOutput(
            path=file_path,
            name=f"Precision versus recall of significant {motifs_name} on the {dataset_type}",
        )

    @staticmethod
    def write_output_tables(report_obj, training_plotting_data, test_plotting_data, training_combined_precision, test_combined_precision, motifs_name="motifs", file_suffix=""):
        results_table_name = f"Confusion matrix and precision/recall scores for significant {motifs_name}" + " on the {} set"
        combined_precision_table_name = f"Combined precision scores of {motifs_name}" + " on the {} set for each TP value on the " + str(report_obj.training_set_name)

        train_results_table = report_obj._write_output_table(training_plotting_data, report_obj.result_path / f"training_set_scores{file_suffix}.csv", results_table_name.format(report_obj.training_set_name))
        test_results_table = report_obj._write_output_table(test_plotting_data, report_obj.result_path / f"test_set_scores{file_suffix}.csv", results_table_name.format(report_obj.test_set_name))
        training_combined_precision_table = report_obj._write_output_table(training_combined_precision, report_obj.result_path / f"training_combined_precision{file_suffix}.csv", combined_precision_table_name.format(report_obj.training_set_name))
        test_combined_precision_table = report_obj._write_output_table(test_combined_precision, report_obj.result_path / f"test_combined_precision{file_suffix}.csv", combined_precision_table_name.format(report_obj.test_set_name))

        return [table for table in [train_results_table, test_results_table, training_combined_precision_table, test_combined_precision_table] if table is not None]

    @staticmethod
    def write_plots(report_obj, training_plotting_data, test_plotting_data, training_combined_precision, test_combined_precision, training_tp_cutoff, test_tp_cutoff, motifs_name="motifs", file_suffix=""):
        training_tp_plot = report_obj._safe_plot(plot_callable="_plot_precision_per_tp", plotting_data=training_plotting_data, combined_precision=training_combined_precision, dataset_type=report_obj.training_set_name, file_path=report_obj.result_path / f"training_precision_per_tp{file_suffix}.html", motifs_name=motifs_name, tp_cutoff=training_tp_cutoff)
        test_tp_plot = report_obj._safe_plot(plot_callable="_plot_precision_per_tp", plotting_data=test_plotting_data, combined_precision=test_combined_precision, dataset_type=report_obj.test_set_name, file_path=report_obj.result_path / f"test_precision_per_tp{file_suffix}.html", motifs_name=motifs_name, tp_cutoff=test_tp_cutoff)
        training_pr_plot = report_obj._safe_plot(plot_callable="_plot_precision_recall", plotting_data=training_plotting_data, dataset_type=report_obj.training_set_name, file_path=report_obj.result_path / f"training_precision_recall{file_suffix}.html", motifs_name=motifs_name)
        test_pr_plot = report_obj._safe_plot(plot_callable="_plot_precision_recall", plotting_data=test_plotting_data, dataset_type=report_obj.test_set_name, file_path=report_obj.result_path / f"test_precision_recall{file_suffix}.html", motifs_name=motifs_name)

        return [plot for plot in [training_tp_plot, test_tp_plot, training_pr_plot, test_pr_plot] if plot is not None]
